------------------------------------------------------------------------------
--                                                                          --
--                         BlackShield COMPONENTS                           --
--                                                                          --
--                           E B N F . L E X E R                            --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                     Copyright (C) 2010, J. Hugues.                       --
--                                                                          --
-- BlackShield is free software;  you  can  redistribute  it and/or  modify --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. BlackShield is distributed  in  the  hope  that it will be--
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with BlackShield; see file COPYING--
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--   BlackShield is maintained by Jerome Hugues <hugues.jerome@gmail.com>   --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Characters.Latin_1;
with System;                   use type System.Address;

with Spark.Character_Handling;
with Spark.OS_Lib;

with Errors;
with Namet;
with Types;                    use type Types.Int, Types.Byte, Types.Name_Id;

package body EBNF.Lexer is

   Initialized : Boolean := False;

   type Token_Image_Array is array (Token_Type) of Types.Name_Id;
   Token_Image : Token_Image_Array :=
      Token_Image_Array'(Token_Type => Types.No_Name);

   ----------------------
   -- Token_Type_Image --
   ----------------------

   function Token_Type_Image
     (T    : Token_Type)
      return Spark.Strings.SparkString
   is
      Result : Spark.Strings.SparkString;
      The_Image : Types.Name_Id;

   begin
      The_Image := Token_Image (T);

      if The_Image = Types.No_Name then
         Result := Spark.Strings.EmptyString;
      else
         --# assert The_Image > Types.No_Name;
         Result := Namet.Name_String (The_Image);
      end if;

      return Result;
   end Token_Type_Image;

   ---------------
   -- Load_File --
   ---------------

   procedure Load_File (Source_Name : Types.Name_Id) is
      Source_File : System.Address;
      Status : Integer;
      Ch : Integer;
   begin
      Source_File := Spark.OS_Lib.FOpen_Read (Namet.Name_String (Source_Name));
      if Source_File = System.Null_Address then
         Errors.Display_Error ("I/O Error: file error");

      else
         Token_Location.Scan := 1;

         loop
            Ch := Spark.OS_Lib.fgetc (Source_File);
            exit when Spark.OS_Lib.feof (Source_File) /= 0
              or else Token_Location.Scan >= Types.Text_Ptr'Last - 1;
            --# assert Token_Location.Scan < Types.Text_Ptr'Last - 1
            --#    and Ch >= 0
            --#    and Ch <= 255;
            Buffer (Token_Location.Scan) := Character'Val (Ch);
            Token_Location.Scan := Token_Location.Scan + 1;
         end loop;
         Buffer (Token_Location.Scan + 1) := Types.EOF;

         Status := Spark.OS_Lib.fclose (Source_File);
         if Status /= 0 then
            Errors.Display_Error ("I/O Error: file error");
         end if;
      end if;
   end Load_File;

   --------------
   -- New_Line --
   --------------

   procedure New_Line is
   begin
      if Token_Location.Scan < Types.Text_Ptr_Max
        and then Token_Location.Line < Types.Int'Last
      then
         Token_Location.Scan      := Token_Location.Scan + 1;
         Token_Location.First_Pos := Token_Location.Scan;
         Token_Location.Last_Pos  := Token_Location.Scan;
         Token_Location.Line      := Token_Location.Line + 1;
      else
         --   XXX
         null;
      end if;
   end New_Line;

   ---------------
   -- New_Token --
   ---------------

   procedure New_Token (Token : Token_Type; Token_Img : String) is
   begin
      Namet.Set_Str_To_Name_Buffer (Token_Img);
      Namet.Name_Find (Token_Image (Token));

      if Token in Keyword_Type then
         Namet.Set_Name_Table_Byte
           (Token_Image (Token),
            Types.Byte (Token_Type'Pos (Token)));
      end if;
   end New_Token;

   -----------------
   -- Skip_Spaces --
   -----------------

   procedure Skip_Spaces is
      C         : Character;
      Must_Exit : Boolean := False;
   begin
      loop
         C := Buffer (Token_Location.Scan);
         case C is
            when ' ' | Ada.Characters.Latin_1.HT =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               else
                  Must_Exit := True;
               end if;
            when Ada.Characters.Latin_1.LF |
                 Ada.Characters.Latin_1.FF |
                 Ada.Characters.Latin_1.CR |
                 Ada.Characters.Latin_1.VT =>
               New_Line;
            when others =>
               Must_Exit := True;
         end case;

         exit when Must_Exit;
      end loop;
   end Skip_Spaces;

   --------------
   -- To_Token --
   --------------

   procedure To_Token (Name : Types.Name_Id; T : out Token_Type) is
      B          : Types.Byte;
      Result     : Token_Type := T_Error;
      Token_Name : Types.Name_Id;

   begin
      Namet.Get_Name_String (Name);
      if Namet.Name_Len > 0 then
         for J in Positive range 1 .. Namet.Name_Len loop
            Namet.Name_Buffer (J) :=
              Spark.Character_Handling.To_Lower (Namet.Name_Buffer (J));
         end loop;
         Namet.Name_Find (Token_Name);
         B := Namet.Get_Name_Table_Byte (Token_Name);
         if B <= Last_Token_Pos then
            Result := Token_Type'Val (B);
         end if;

         T := Result;
      else
         T := T_Error;
      end if;
   end To_Token;

   ---------------------
   -- Scan_Identifier --
   ---------------------

   procedure Scan_Identifier (Fatal : Boolean) is
   begin
      --  Read identifier

      Namet.Name_Len := 0;
      while Buffer (Token_Location.Scan) /= '''
        and then Buffer (Token_Location.Scan) /= '='
        and then Buffer (Token_Location.Scan) /= '-'
        and then Buffer (Token_Location.Scan) /= ';'
        and then Buffer (Token_Location.Scan) /= '|'
        and then Buffer (Token_Location.Scan) /= '}'
        and then Buffer (Token_Location.Scan) /= '{'
        and then Buffer (Token_Location.Scan) /= '['
        and then Buffer (Token_Location.Scan) /= ']'
        and then Buffer (Token_Location.Scan) /= '('
        and then Buffer (Token_Location.Scan) /= ')'
        and then Buffer (Token_Location.Scan) /= ','
        and then Buffer (Token_Location.Scan) /= ' '
        and then Token_Location.Scan < Types.Text_Ptr'Last
        and then Buffer (Token_Location.Scan) /= Types.EOF
      loop
         --# assert Token_Location.Scan < Types.Text_Ptr'Last
         --# and Namet.Name_Len >= 0;

         if Buffer (Token_Location.Scan) /= Ada.Characters.Latin_1.HT
           and then Buffer (Token_Location.Scan) /=
                    Ada.Characters.Latin_1.CR
           and then Buffer (Token_Location.Scan) /=
                    Ada.Characters.Latin_1.LF
         then
            if Namet.Name_Len < Namet.Name_Buffer_Max then
               Namet.Name_Len                     := Namet.Name_Len + 1;
               Namet.Name_Buffer (Namet.Name_Len) :=
                 Buffer (Token_Location.Scan);
            else
               Errors.Set_Error_Loc (1, Token_Location);
               Errors.Display_Error ("ICE");
            end if;
         else
            New_Line; --  XXX
         end if;

         if Token_Location.Scan < Types.Text_Ptr_Max then
            Token_Location.Scan := Token_Location.Scan + 1;
         end if;
      end loop;

      if Namet.Name_Len = 0 then
         if Fatal then
            Errors.Set_Error_Loc (1, Token_Location);

            --  Note: we expect the assignment to Namet.Name_Len to be
            --  ineffective: we are trashing the value of Namet.Name_Len
            --  after the call to Errors.Display_Error

            --# accept F, 10, Namet.Name_Len,
            --#  "Expected ineffective assignment to Namet.Name_Len";
            Errors.Display_Error
              ("identifier must start with alphabetic character");
         end if;
         Namet.Name_Buffer (1) := ' ';
         Namet.Name_Len        := 1;

      else
         Namet.Name_Find (Token_Name);
         To_Token (Token_Name, Token);
         if Token = T_Error then
            Token := T_Identifier;
         end if;
      end if;
   end Scan_Identifier;

   ----------------
   -- Scan_Token --
   ----------------

   procedure Scan_Token (Fatal : Boolean) is
   begin
      if Token_Location.Scan = Types.Text_Ptr_Max then
         Token := T_EOF;

      elsif Token /= T_EOF then
         --# assert Token_Location.Scan < Types.Text_Ptr_Max;

         Token := T_Error;

         loop
            exit when Token /= T_Error;

            Skip_Spaces;

            --# assert Token_Location.Scan <= Types.Text_Ptr_Max;

            Token_Location.Last_Pos := Token_Location.Scan;

            case Buffer (Token_Location.Scan) is
               when Ada.Characters.Latin_1.LF
                 | Ada.Characters.Latin_1.FF
                 | Ada.Characters.Latin_1.CR
                 | Ada.Characters.Latin_1.VT =>
               New_Line;

            when ';' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_Terminator_Symbol;
               Token_Name          := Token_Image (Token);

            when ''' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_First_Quote_Symbol;
               Token_Name          := Token_Image (Token);

            when '-' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_Except_Symbol;
               Token_Name          := Token_Image (Token);

            when '"' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_Second_Quote_Symbol;
               Token_Name          := Token_Image (Token);

            when '|' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_Vertical_Line;
               Token_Name          := Token_Image (Token);

            when '{' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_Left_Curly_Bracket;
               Token_Name          := Token_Image (Token);

            when '}' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_Right_Curly_Bracket;
               Token_Name          := Token_Image (Token);

            when '[' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_Left_Square_Bracket;
               Token_Name          := Token_Image (Token);

            when ']' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_Right_Square_Bracket;
               Token_Name          := Token_Image (Token);

            when ',' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_Comma;
               Token_Name          := Token_Image (Token);

            when '*' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_Asterisk;
               Token_Name          := Token_Image (Token);

            when '=' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_Defining_Symbol;
               Token_Name          := Token_Image (Token);

            when ')' =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_Right_Parenthesis;
               Token_Name          := Token_Image (Token);

            when '(' =>
               --  Comment like (* ... *) (not nested)

               if Token_Location.Scan < Types.Text_Ptr_Max - 1
                 and then Buffer (Token_Location.Scan + 1) = '*' then
                  Token_Location.Scan := Token_Location.Scan + 2;
                  while Buffer (Token_Location.Scan) /= Types.EOF
                    and then Token_Location.Scan < Types.Text_Ptr_Max
                    and then (Buffer (Token_Location.Scan) /= '*'
                             or else Buffer (Token_Location.Scan + 1) /=
                                     ')')
                  loop
                     case Buffer (Token_Location.Scan) is
                        when Ada.Characters.Latin_1.LF |
                             Ada.Characters.Latin_1.FF |
                             Ada.Characters.Latin_1.CR |
                             Ada.Characters.Latin_1.VT =>
                           New_Line;
                        when others =>
                           Token_Location.Scan := Token_Location.Scan + 1;
                     end case;
                  end loop;

                  if Buffer (Token_Location.Scan) = Types.EOF then
                     Errors.Set_Error_Loc (1, Token_Location);
                     Errors.Display_Error ("unterminated comment");
                  end if;

                  if Token_Location.Scan < Types.Text_Ptr_Max - 1 then
                     --  Skip char sequence *)
                     Token_Location.Scan := Token_Location.Scan + 2;
                  else
                     Errors.Set_Error_Loc (1, Token_Location);
                     Errors.Display_Error ("unterminated comment");
                  end if;

               else
                  if Token_Location.Scan < Types.Text_Ptr_Max then
                     Token_Location.Scan := Token_Location.Scan + 1;
                  end if;
                  Token               := T_Left_Parenthesis;
               end if;

            when Types.EOF =>
               if Token_Location.Scan < Types.Text_Ptr_Max then
                  Token_Location.Scan := Token_Location.Scan + 1;
               end if;
               Token               := T_EOF;

            when others =>
               if Spark.Character_Handling.Is_Letter
                    (Buffer (Token_Location.Scan))
               then
                  Scan_Identifier (Fatal);

               else
                  Errors.Set_Error_Loc (1, Token_Location);
                  Errors.Display_Error ("Invalid character");
               end if;
            end case;
         end loop;
      end if;
   end Scan_Token;

   ----------------
   -- Next_Token --
   ----------------

   procedure Next_Token (The_Token : out Token_Type) is
      Current_Token_Name     : Types.Name_Id;
      Current_Token          : Token_Type;
      Current_Token_Location : Locations.Location;
      Next_Token_Value       : Token_Type;
   begin
      Current_Token_Name     := Token_Name;
      Current_Token          := Token;
      Current_Token_Location := Token_Location;

      --  Note: we expect the assignment to be ineffective: we are
      --  trashing these two values to restore the previous value.

      --# accept F, 10, Token_Name,
      --#  "Expected ineffective assignment to Token_Name" &
      --#        F, 10, Token_Location,
      --#  "Expected ineffective assignment to Token_Location";

      Scan_Token (Fatal => False);
      Next_Token_Value := Token;
      Token_Name       := Current_Token_Name;
      Token            := Current_Token;
      Token_Location   := Current_Token_Location;

      The_Token := Next_Token_Value;
   end Next_Token;

   -------------
   -- Process --
   -------------

   procedure Process (Source_Name : Types.Name_Id) is
   begin
      if not Initialized then
         Initialized := True;

         --  Enter all the alphabetic keywords in the name table

         --  Note: we expect the assignment to Namet.Name_Len to be
         --  ineffective: we are trashing the value of Namet.Name_Len
         --  after each call to New_Token.

         --# accept F, 10, Namet.Name_Len,
         --#  "Expected ineffective assignment to Namet.Name_Len";
         New_Token (T_Error, "<error>");
         New_Token (T_Identifier, "<identifier>");
         New_Token (T_First_String, "<first string>");
         New_Token (T_EOF, "<end of file>");
         New_Token (T_Comma, ",");
         New_Token (T_Defining_Symbol, "=");
         New_Token (T_Vertical_Line, "|");
         New_Token (T_Divide_Symbol, "/");
         New_Token (T_Exclamation_Mark, "!");
         New_Token (T_Asterisk_Right_Parenthesis, "*)");
         New_Token (T_Right_Parenthesis, ")");
         New_Token (T_Right_Square_Bracket, "]");
         New_Token (T_Right_Curly_Bracket, "}");
         New_Token (T_Hyphen_Minus, "_");
         New_Token (T_First_Quote_Symbol, "'");
         New_Token (T_Asterisk, "*");
         New_Token (T_Except_Symbol, "-");
         New_Token (T_Second_Quote_Symbol, """");
         New_Token (T_Question_Mark, "?");
         New_Token (T_Left_Parenthesis_Asterisk, "(*");
         New_Token (T_Left_Parenthesis, "(");
         New_Token (T_Left_Square_Bracket, "[");
         New_Token (T_Left_Curly_Bracket, "{");
         New_Token (T_Period, ".");
         New_Token (T_Terminator_Symbol, ";");
         --# end accept;
      end if;

      --  Load source file in a buffer

      Load_File (Source_Name);

      --  Reset at the beginning

      Token_Location.Scan      := 1;
      Token_Location.First_Pos := 1;
      Token_Location.Last_Pos  := 1;
      Token_Location.Line      := 1;
      Token_Location.Base_Name := Source_Name;
   end Process;

   -------------------
   -- Restore_Lexer --
   -------------------

   procedure Restore_Lexer (State : Locations.Location) is
   begin
      Token_Location := State;
   end Restore_Lexer;

   ----------------
   -- Save_Lexer --
   ----------------

   procedure Save_Lexer (State : out Locations.Location) is
   begin
      State := Token_Location;
   end Save_Lexer;

   -------------------------------
   -- Scan_String_Literal_Value --
   -------------------------------

   procedure Scan_String_Literal_Value (Stop_Character : Character) is
      Ch : Character;

   begin
      Namet.Name_Len := 0;   --  initialize string buffer
      loop
         Ch := Buffer (Token_Location.Scan);
         if Ch /= Stop_Character
           and then Token_Location.Scan < Buffer'Last
         then
            Namet.Add_Char_To_Name_Buffer (Ch);
            Token_Location.Scan := Token_Location.Scan + 1;
         end if;
         exit when Ch = Stop_Character
           or else Token_Location.Scan = Buffer'Last;
      end loop;

      if Namet.Name_Len = 0 then
         Token_Name := Types.No_Name;
      else
         Namet.Name_Find (Token_Name);
      end if;
   end Scan_String_Literal_Value;

   --------------------
   -- Scan_The_Token --
   --------------------

   procedure Scan_The_Token (T : Token_Type) is
      Loc : Locations.Location;
   begin
      Loc := Token_Location;

      if T = T_First_String then
         Scan_String_Literal_Value (''');
         Token := T_First_String;

      elsif T = T_Second_String then
         Scan_String_Literal_Value ('"');
         Token := T_Second_String;
      else
         Scan_Token (Fatal => True);
      end if;

      if T = T_Identifier then
         Token := T_Identifier;
      end if;

      if T /= Token then
         if T = T_Terminator_Symbol then
            Loc.Last_Pos := Loc.Scan;
         else
            Loc := Token_Location;
         end if;

         Errors.Set_Error_Loc (1, Loc);
         Errors.Set_Error_Name (1, Token_Image (T));
         Errors.Display_Error ("Expected token #");
         Token := T_Error;
      end if;
   end Scan_The_Token;

end EBNF.Lexer;
