------------------------------------------------------------------------------
--                                                                          --
--                         BlackShield COMPONENTS                           --
--                                                                          --
--                           E B N F . L E X E R                            --
--                                                                          --
--                                 S p e c                                  --
--                                                                          --
--                     Copyright (C) 2010, J. Hugues.                       --
--                                                                          --
-- BlackShield is free software;  you  can  redistribute  it and/or  modify --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. BlackShield is distributed  in  the  hope  that it will be--
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with BlackShield; see file COPYING--
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--   BlackShield is maintained by Jerome Hugues <hugues.jerome@gmail.com>   --
--                                                                          --
------------------------------------------------------------------------------

with Locations;
with Types;
with Spark.Strings;
--# inherit Ada.Characters.Latin_1,
--#         Errors,
--#         Locations,
--#         Namet,
--#         Output,
--#         Spark.Character_Handling,
--#         Spark.OS_Lib,
--#         Spark.Strings,
--#         Types,
--# System;

package EBNF.Lexer
   --# own Buffer,
   --#     Initialized,
   --#     Token,
   --#     Token_Image,
   --#     Token_Location,
   --#     Token_Name;
   --# initializes Buffer,
   --#             Initialized,
   --#             Token,
   --#             Token_Image,
   --#             Token_Location,
   --#             Token_Name;
is

   type Token_Type is (
      T_Error,
      T_Identifier,
      T_First_String,
      T_Second_String,

   --  EBNF terminal characters

      T_Comma,
      T_Defining_Symbol,
      T_Vertical_Line,
      T_Divide_Symbol,
      T_Exclamation_Mark,
      T_Asterisk_Right_Parenthesis,
      T_Right_Parenthesis,
      T_Right_Square_Bracket,
      T_Right_Curly_Bracket,
      T_Hyphen_Minus,
      T_First_Quote_Symbol,
      T_Second_Quote_Symbol,
      T_Asterisk,
      T_Except_Symbol,
      T_Quotation_Mark,
      T_Question_Mark,
      T_Left_Parenthesis_Asterisk,
      T_Left_Parenthesis,
      T_Left_Square_Bracket,
      T_Left_Curly_Bracket,
      T_Period,
      T_Terminator_Symbol,
      T_EOF);

   First_Token_Pos : constant := Token_Type'Pos (Token_Type'First);
   Last_Token_Pos  : constant := Token_Type'Pos (Token_Type'Last);

   subtype Token_Range is Natural range First_Token_Pos .. Last_Token_Pos;
   type Token_List_Type is array (Token_Range) of Token_Type;

   subtype Keyword_Type is Token_Type range
      T_Defining_Symbol .. T_Terminator_Symbol;

   Token          : Token_Type         := T_Error;
   Token_Name     : Types.Name_Id      := Types.No_Name;
   Token_Location : Locations.Location := Locations.No_Location;

   Buffer : Types.Text_Buffer := Types.Text_Buffer'(Types.Text_Ptr => ' ');

   function Token_Type_Image
     (T    : Token_Type)
      return Spark.Strings.SparkString;
   --# global in Namet.State;
   --#        in Token_Image;
   --  Return an image of token T.

   procedure Scan_Token (Fatal : Boolean);
   --# global in     Buffer;
   --#        in     Token_Image;
   --#        in out Errors.Error_Int;
   --#        in out Errors.Error_Loc;
   --#        in out Errors.Error_Name;
   --#        in out Errors.N_Errors;
   --#        in out Namet.Name_Buffer;
   --#        in out Namet.Name_Len;
   --#        in out Namet.State;
   --#        in out Output.State;
   --#        in out Token;
   --#        in out Token_Location;
   --#        in out Token_Name;
   --  Scan token and update global variables Token, Token_Name (for
   --  identifiers and literals) and Token_Location.  If Fatal is
   --  false, do not report any error and do not fail on minor errors
   --  like detecting a string which appears to be a wide string.

   procedure Scan_The_Token (T : Token_Type);
   --# global in     Buffer;
   --#        in     Token_Image;
   --#        in out Errors.Error_Int;
   --#        in out Errors.Error_Loc;
   --#        in out Errors.Error_Name;
   --#        in out Errors.N_Errors;
   --#        in out Namet.Name_Buffer;
   --#        in out Namet.Name_Len;
   --#        in out Namet.State;
   --#        in out Output.State;
   --#        in out Token;
   --#        in out Token_Location;
   --#        in out Token_Name;
   --  Same as above. When the current token is not the expected token
   --  T, an error message is output and Token is set to T_Error. As a
   --  special case, when T_Semi_Colon is expected, we output an error
   --  location at the end of the line instead of the current token
   --  location.

   procedure Next_Token (The_Token : out Token_Type);
   --# global in     Buffer;
   --#        in     Token_Image;
   --#        in out Errors.Error_Int;
   --#        in out Errors.Error_Loc;
   --#        in out Errors.Error_Name;
   --#        in out Errors.N_Errors;
   --#        in out Namet.Name_Buffer;
   --#        in out Namet.Name_Len;
   --#        in out Namet.State;
   --#        in out Output.State;
   --#        in out Token;
   --#        in out Token_Location;
   --#        in out Token_Name;
   --  Return next token but do not update the lexer state that is
   --  Token, Token_Name and Token_Location.

   procedure Save_Lexer (State : out Locations.Location);
   --# global in Token_Location;
   --  Saves the current location in the State variable

   procedure Restore_Lexer (State : Locations.Location);
   --# global out Token_Location;
   --  Modifies the current location in the IDL specification

   procedure Process (Source_Name : Types.Name_Id);
   --# global in out Buffer;
   --#        in out Errors.Error_Int;
   --#        in out Errors.Error_Loc;
   --#        in out Errors.Error_Name;
   --#        in out Errors.N_Errors;
   --#        in out Initialized;
   --#        in out Namet.Name_Buffer;
   --#        in out Namet.Name_Len;
   --#        in out Namet.State;
   --#        in out Output.State;
   --#        in out Token_Image;
   --#        in out Token_Location;
   --# pre Source_Name > Types.No_Name;
   --  Load file Source in the lexer

private

   procedure Load_File (Source_Name : Types.Name_Id);
   --# global in out Buffer;
   --#        in out Errors.Error_Int;
   --#        in out Errors.Error_Loc;
   --#        in out Errors.Error_Name;
   --#        in out Errors.N_Errors;
   --#        in out Namet.Name_Buffer;
   --#        in out Namet.Name_Len;
   --#        in out Namet.State;
   --#        in out Output.State;
   --#        in out Token_Location;
   --# pre Source_Name > Types.No_Name;
   --  Loads a file in the buffer and then closes it.

   procedure Scan_Identifier (Fatal : Boolean);
   --# global in     Buffer;
   --#        in out Errors.Error_Int;
   --#        in out Errors.Error_Loc;
   --#        in out Errors.Error_Name;
   --#        in out Errors.N_Errors;
   --#        in out Namet.Name_Buffer;
   --#        in out Namet.State;
   --#        in out Output.State;
   --#        in out Token;
   --#        in out Token_Location;
   --#        in out Token_Name;
   --#           out Namet.Name_Len;

   procedure New_Token (Token : Token_Type; Token_Img : String);
   --# global in out Namet.Name_Buffer;
   --#        in out Namet.State;
   --#        in out Token_Image;
   --#           out Namet.Name_Len;
   --  Evaluate token image and store it in Token_Image table. When
   --  Token is a graphical character, embrace its image between
   --  single quotes ('<<' and '>>' are considered as graphical
   --  characters). When Token is a keyword, embrace its image between
   --  double quotes.

   procedure New_Line;
   --# global in out Token_Location;
   --  Increment the line number and save the current position in the
   --  buffer in order to compute later on the column number.

   procedure Skip_Spaces;
   --# global in     Buffer;
   --#        in out Token_Location;
   --  Skip all spaces

   procedure To_Token (Name : Types.Name_Id; T : out Token_Type);
   --# global in out Namet.Name_Buffer;
   --#        in out Namet.State;
   --#           out Namet.Name_Len;
   --# pre Name > Types.No_Name;
   --  Return the token matching Name. Otherwise, return T_Error.

   procedure Scan_String_Literal_Value (Stop_Character : Character);
   --# global in     Buffer;
   --#        in out Namet.Name_Buffer;
   --#        in out Namet.State;
   --#        in out Token_Location;
   --#           out Namet.Name_Len;
   --#           out Token_Name;

end EBNF.Lexer;
