------------------------------------------------------------------------------
--                                                                          --
--                         BlackShield COMPONENTS                           --
--                                                                          --
--                         E B N F . P R I N T E R                          --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                     Copyright (C) 2010, J. Hugues.                       --
--                                                                          --
-- BlackShield is free software;  you  can  redistribute  it and/or  modify --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. BlackShield is distributed  in  the  hope  that it will be--
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with BlackShield; see file COPYING--
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--   BlackShield is maintained by Jerome Hugues <hugues.jerome@gmail.com>   --
--                                                                          --
------------------------------------------------------------------------------

with EBNF.Lexer;
with Namet;
with Output;
with EBNF.Nodes;
with Types;      use type Types.Node_Id;

package body EBNF.Printer
   --# own Stack_State is Stack.State;
is

   --# inherit Types;
   package Stack
      --# own State;
      --# initializes State;
   is

      type Element_Kind_Type is (
         E_No_Element,

      --  Terminals

         E_Concatenate_Symbol,
         E_Definition_Separator_Symbol,
         E_Defining_Symbol,
         E_Terminator_Symbol,
         E_Start_Group_Symbol,
         E_End_Group_Symbol,
         E_Start_Option_Symbol,
         E_End_Option_Symbol,
         E_Start_Repeat_Symbol,
         E_End_Repeat_Symbol,

      --  Non terminal

         E_Composite);

      function Is_Empty return Boolean;
      --# global in State;

      procedure Push
        (Element_Kind : Element_Kind_Type;
         A_Element    : Types.Node_Id);
      --# global in out State;

      procedure Pop
        (Element_Kind : out Element_Kind_Type;
         A_Element    : out Types.Node_Id);
      --# global in out State;

      procedure Push_Temp
        (Element_Kind : Element_Kind_Type;
         A_Element    : Types.Node_Id);
      --# global in out State;

      procedure Pump;
      --# global in out State;

   private

      procedure Pop_Temp
        (Element_Kind : out Element_Kind_Type;
         A_Element    : out Types.Node_Id);
      --# global in out State;

   end Stack;

   package body Stack
      --# own State is The_Stack, Stack_Pointer, Temp_Stack,
      --# Temp_Stack_Pointer;
   is

      type Stack_Element is record
         Element_Kind : Element_Kind_Type;
         A_Element    : Types.Node_Id;
      end record;

      Null_Element : constant Stack_Element :=
         Stack_Element'
        (Element_Kind => E_No_Element,
         A_Element    => Types.No_Node);

      Stack_Size : constant := 1_000;
      subtype Stack_Pointer_Type is Integer range 0 .. Stack_Size;
      subtype Index is Integer range 1 .. Stack_Pointer_Type'Last;
      type Stack_Type is array (Index) of Stack_Element;

      The_Stack     : Stack_Type         :=
         Stack_Type'(Index => Null_Element);
      Stack_Pointer : Stack_Pointer_Type := 0;

      Temp_Stack         : Stack_Type         :=
         Stack_Type'(Index => Null_Element);
      Temp_Stack_Pointer : Stack_Pointer_Type := 0;

      --------------
      -- Is_Empty --
      --------------

      function Is_Empty return Boolean
      --# global in Stack_Pointer;
      is
      begin
         return Stack_Pointer = 0;
      end Is_Empty;

      ----------
      -- Push --
      ----------

      procedure Push
        (Element_Kind : Element_Kind_Type;
         A_Element    : Types.Node_Id)
         --# global in out Stack_Pointer;
         --#        in out The_Stack;
      is
      begin
         if Stack_Pointer < The_Stack'Last then
            Stack_Pointer             := Stack_Pointer + 1;
            The_Stack (Stack_Pointer) :=
              Stack_Element'
              (Element_Kind => Element_Kind,
               A_Element    => A_Element);
         end if;
      end Push;

      ---------------
      -- Push_Temp --
      ---------------

      procedure Push_Temp
        (Element_Kind : Element_Kind_Type;
         A_Element    : Types.Node_Id)
         --# global in out Temp_Stack;
         --#        in out Temp_Stack_Pointer;
      is
      begin
         if Temp_Stack_Pointer < Temp_Stack'Last then
            Temp_Stack_Pointer              := Temp_Stack_Pointer + 1;
            Temp_Stack (Temp_Stack_Pointer) :=
              Stack_Element'
              (Element_Kind => Element_Kind,
               A_Element    => A_Element);
         end if;
         --  XXX
      end Push_Temp;

      ---------
      -- Pop --
      ---------

      procedure Pop
        (Element_Kind : out Element_Kind_Type;
         A_Element    : out Types.Node_Id)
         --# global in     The_Stack;
         --#        in out Stack_Pointer;
      is
      begin
         if Stack_Pointer = 0 then
            Element_Kind := E_No_Element;
            A_Element    := Types.No_Node;

         else
            Element_Kind  := The_Stack (Stack_Pointer).Element_Kind;
            A_Element     := The_Stack (Stack_Pointer).A_Element;
            Stack_Pointer := Stack_Pointer - 1;
         end if;
      end Pop;

      --------------
      -- Pop_Temp --
      --------------

      procedure Pop_Temp
        (Element_Kind : out Element_Kind_Type;
         A_Element    : out Types.Node_Id)
         --# global in     Temp_Stack;
         --#        in out Temp_Stack_Pointer;
      is
      begin
         if Temp_Stack_Pointer = 0 then
            Element_Kind := E_No_Element;
            A_Element    := Types.No_Node;

         else
            Element_Kind       :=
              Temp_Stack (Temp_Stack_Pointer).Element_Kind;
            A_Element          := Temp_Stack (Temp_Stack_Pointer).A_Element;
            Temp_Stack_Pointer := Temp_Stack_Pointer - 1;
         end if;
      end Pop_Temp;

      ---------
      -- Pump --
      ---------

      procedure Pump
         --# global in     Temp_Stack;
         --#        in out Stack_Pointer;
         --#        in out Temp_Stack_Pointer;
         --#        in out The_Stack;
      is
         The_Element_Kind : Element_Kind_Type;
         The_Element      : Types.Node_Id;
      begin
         loop
            Pop_Temp (The_Element_Kind, The_Element);
            exit when The_Element_Kind = E_No_Element;

            Push (The_Element_Kind, The_Element);
         end loop;
      end Pump;
   end Stack;

   -----------------------------------
   -- W_Definition_Separator_Symbol --
   -----------------------------------

   procedure W_Definition_Separator_Symbol is
   begin
      Output.Write_SparkStr (Lexer.Token_Type_Image (Lexer.T_Vertical_Line));
   end W_Definition_Separator_Symbol;

   --------------------------
   -- W_Concatenate_Symbol --
   --------------------------

   procedure W_Concatenate_Symbol is
   begin
      Output.Write_SparkStr (Lexer.Token_Type_Image (Lexer.T_Comma));
   end W_Concatenate_Symbol;

   -----------------------
   -- W_Defining_Symbol --
   -----------------------

   --  defining symbol = '=';

   procedure W_Defining_Symbol is
   begin
      Output.Write_SparkStr
        (Lexer.Token_Type_Image (Lexer.T_Defining_Symbol));
   end W_Defining_Symbol;

   -------------------------
   -- W_Terminator_Symbol --
   -------------------------

   --  terminator symbol = ';' | '.';

   procedure W_Terminator_Symbol is
   begin
      Output.Write_SparkStr
        (Lexer.Token_Type_Image (Lexer.T_Terminator_Symbol));
      Output.Write_Eol;
   end W_Terminator_Symbol;

   --     ---------------------
   --     -- W_Except_Symbol --
   --     ---------------------

   --     procedure W_Except_Symbol is
   --     begin
   --     Output.Write_SparkStr (Lexer.Token_Type_Image
   --(Lexer.T_Except_Symbol));
   --     end W_Except_Symbol;

   --------------------------
   -- W_Start_Group_Symbol --
   --------------------------

   procedure W_Start_Group_Symbol is
   begin
      Output.Write_SparkStr
        (Lexer.Token_Type_Image (Lexer.T_Left_Parenthesis));
   end W_Start_Group_Symbol;

   ------------------------
   -- W_End_Group_Symbol --
   ------------------------

   procedure W_End_Group_Symbol is
   begin
      Output.Write_SparkStr
        (Lexer.Token_Type_Image (Lexer.T_Right_Parenthesis));
   end W_End_Group_Symbol;

   ---------------------------
   -- W_Start_Option_Symbol --
   ---------------------------

   --  start option symbol = '[' | '(/';
   --  XXX incomplete

   procedure W_Start_Option_Symbol is
   begin
      Output.Write_SparkStr
        (Lexer.Token_Type_Image (Lexer.T_Left_Square_Bracket));
   end W_Start_Option_Symbol;

   -------------------------
   -- W_End_Option_Symbol --
   -------------------------

   --  end option symbol = ']' | '/)';
   --  XXX incomplete

   procedure W_End_Option_Symbol is
   begin
      Output.Write_SparkStr
        (Lexer.Token_Type_Image (Lexer.T_Right_Square_Bracket));
   end W_End_Option_Symbol;

   ---------------------------
   -- W_Start_Repeat_Symbol --
   ---------------------------

   --  start repeat symbol = '{' | '(:';
   --  XXX incomplete

   procedure W_Start_Repeat_Symbol is
   begin
      Output.Write_SparkStr
        (Lexer.Token_Type_Image (Lexer.T_Left_Curly_Bracket));
   end W_Start_Repeat_Symbol;

   -------------------------
   -- W_End_Repeat_Symbol --
   -------------------------

   --  end repeat symbol = '}' | ':)';
   --  XXX incomplete

   procedure W_End_Repeat_Symbol is
   begin
      Output.Write_SparkStr
        (Lexer.Token_Type_Image (Lexer.T_Right_Curly_Bracket));
   end W_End_Repeat_Symbol;

   ---------------------------
   -- W_First_Quote_Symnbol --
   ---------------------------

   --  first quote symbol = "'";

   procedure W_First_Quote_Symbol is
   begin
      Output.Write_SparkStr
        (Lexer.Token_Type_Image (Lexer.T_First_Quote_Symbol));
   end W_First_Quote_Symbol;

   ----------------------------
   -- W_Second_Quote_Symnbol --
   ----------------------------

   --  second quote symbol = '"';

   --   procedure W_Second_Quote_Symbol is
   --   begin
   --      Output.Write_SparkStr
   --  (Lexer.Token_Type_Image (Lexer.T_Second_Quote_Symbol));
   --   end W_Second_Quote_Symbol;

   ------------------
   -- W_Identifier --
   ------------------

   procedure W_Identifier (N : Types.Node_Id) is
   begin
      Output.Write_SparkStr (Namet.Name_String (Nodes.Name (N)));
   end W_Identifier;

   -----------------------
   -- W_Meta_Identifier --
   -----------------------

   --  (* see 4.14 *) meta identifier
   --    = letter, {meta identifier character};

   procedure W_Meta_Identifier (N : Types.Node_Id) is
      The_Identifier : Types.List_Id;
      I              : Types.Node_Id;
   begin
      The_Identifier := Nodes.Identifier (N);
      I              := Nodes.First_Node (The_Identifier);
      while I /= Types.No_Node loop
         W_Identifier (I);
         I := Nodes.Next_Node (I);
         if I /= Types.No_Node then
            Output.Write_Str (" ");
         end if;
      end loop;
   end W_Meta_Identifier;

   --     ---------------------------
   --     -- W_Syntactic_Exception --
   --     ---------------------------

   --     --  (* see 4.7 *) syntactic exception
   --     --   = syntactic factor;

   --     procedure W_Syntactic_Exception (N : Types.Node_Id) is
   --        The_Syntactic_Factor : Types.Node_Id;
   --        Result_Node : Types.Node_Id;
   --        Loc : Location;
   --     begin
   --        Save_Lexer (Loc);
   --        The_Syntactic_Factor := P_Syntactic_Factor;

   --        Result_Node := New_Node (K_Syntactic_Exception, Loc);
   --        Set_Syntactic_Factor (Result_Node, The_Syntactic_Factor);

   --        return Result_Node;
   --     end W_Syntactic_Exception;

   -----------------------
   -- W_Terminal_String --
   -----------------------

   --  (* see 4.16 *) terminal string
   --    = first quote symbol, first terminal character,
   --      {first terminal character},
   --      first quote symbol
   --     | second quote symbol, second terminal character,
   --       {second terminal character},
   --       second quote symbol;

   procedure W_Terminal_String (N : Types.Node_Id) is
   begin
      W_First_Quote_Symbol;
      Output.Write_SparkStr (Namet.Name_String (Nodes.The_String (N)));
      W_First_Quote_Symbol;
   end W_Terminal_String;

   ----------------------
   -- W_Empty_Sequence --
   ----------------------

   --  (* see 4.21 *) empty sequence
   --    = ;

   procedure W_Empty_Sequence is
   begin
      null;
   end W_Empty_Sequence;

   ------------------------
   -- W_Syntactic_Factor --
   ------------------------

   --  (* see 4.8 *) syntactic factor
   --    = [integer, repetition symbol],
   --      syntactic primary;
   --  XXX incomplete

   procedure W_Syntactic_Factor (N : Types.Node_Id)
      --# global in     Nodes.State;
      --#        in out Stack.State;
      --# pre N > Types.No_Node;
   is
      The_Syntactic_Primary : Types.Node_Id;
   begin
      The_Syntactic_Primary := Nodes.Syntactic_Primary (N);
      Stack.Push (Stack.E_Composite, The_Syntactic_Primary);
   end W_Syntactic_Factor;

   -------------------------
   -- W_Single_Definition --
   -------------------------

   --  (* see 4.5 *) single definition
   --    = syntactic term,
   --      {concatenate symbol, syntactic term};

   procedure W_Single_Definition (N : Types.Node_Id)
      --# global in     Nodes.State;
      --#        in out Stack.State;
      --# pre N > Types.No_Node;
   is
      The_Syntactic_Term : Types.List_Id;
      I                  : Types.Node_Id;
   begin
      The_Syntactic_Term := Nodes.Syntactic_Term (N);
      I                  := Nodes.First_Node (The_Syntactic_Term);

      Stack.Push_Temp (Stack.E_Composite, I);

      I := Nodes.Next_Node (I);
      while I /= Types.No_Node loop
         Stack.Push_Temp (Stack.E_Concatenate_Symbol, Types.No_Node);
         Stack.Push_Temp (Stack.E_Composite, I);
         I := Nodes.Next_Node (I);
      end loop;
      Stack.Pump;
   end W_Single_Definition;

   ------------------------
   -- W_Definitions_List --
   ------------------------

   --  (* see 4.4 *) definitions list
   --    = single definition,
   --      {definition separator symbol,
   --       single definition};

   procedure W_Definitions_List (N : Types.Node_Id)
      --# global in     Nodes.State;
      --#        in out Stack.State;
      --# pre N > Types.No_Node;
   is
      The_Single_Definition : Types.List_Id;
      I                     : Types.Node_Id;
   begin
      The_Single_Definition := Nodes.Single_Definition (N);
      I                     := Nodes.First_Node (The_Single_Definition);

      Stack.Push_Temp (Stack.E_Composite, I);

      I := Nodes.Next_Node (I);
      while I /= Types.No_Node loop
         Stack.Push_Temp (Stack.E_Definition_Separator_Symbol, Types.No_Node);
         Stack.Push_Temp (Stack.E_Composite, I);
         I := Nodes.Next_Node (I);
      end loop;
      Stack.Pump;
   end W_Definitions_List;

   ------------------------
   -- W_Grouped_Sequence --
   ------------------------

   --  (* see 4.13 *) grouped sequence
   --    = start group symbol, definitions list,
   --      end group symbol;

   procedure W_Grouped_Sequence (N : Types.Node_Id)
      --# global in     Nodes.State;
      --#        in out Stack.State;
      --# pre N > Types.No_Node;
   is
   begin
      Stack.Push_Temp (Stack.E_Start_Group_Symbol, Types.No_Node);
      Stack.Push_Temp (Stack.E_Composite, Nodes.Definitions_List (N));
      Stack.Push_Temp (Stack.E_End_Group_Symbol, Types.No_Node);
      Stack.Pump;
   end W_Grouped_Sequence;

   -------------------------
   -- W_Optional_Sequence --
   -------------------------

   --  (* see 4.11 *) optional sequence
   --    = start option symbol, definitions list,
   --      end option symbol;

   procedure W_Optional_Sequence (N : Types.Node_Id)
      --# global in     Nodes.State;
      --#        in out Stack.State;
      --# pre N > Types.No_Node;
   is
   begin
      Stack.Push_Temp (Stack.E_Start_Option_Symbol, Types.No_Node);
      Stack.Push_Temp (Stack.E_Composite, Nodes.Definitions_List (N));
      Stack.Push_Temp (Stack.E_End_Option_Symbol, Types.No_Node);
      Stack.Pump;
   end W_Optional_Sequence;

   -------------------------
   -- W_Repeated_Sequence --
   -------------------------

   --  (* see 4.12 *) repeated sequence
   --    = start repeat symbol, definitions list,
   --      end repeat symbol;

   procedure W_Repeated_Sequence (N : Types.Node_Id)
      --# global in     Nodes.State;
      --#        in out Stack.State;
      --# pre N > Types.No_Node;
   is
   begin
      Stack.Push_Temp (Stack.E_Start_Repeat_Symbol, Types.No_Node);
      Stack.Push_Temp (Stack.E_Composite, Nodes.Definitions_List (N));
      Stack.Push_Temp (Stack.E_End_Repeat_Symbol, Types.No_Node);
      Stack.Pump;
   end W_Repeated_Sequence;

   -------------------------
   -- W_Syntactic_Primary --
   -------------------------

   --  (* see 4.10 *) syntactic primary
   --    = optional sequence
   --      | repeated sequence
   --      | grouped sequence
   --      | meta identifier
   --      | terminal string
   --      | special sequence
   --      | empty sequence;

   procedure W_Syntactic_Primary (N : Types.Node_Id)
      --# global in     Nodes.State;
      --#        in out Stack.State;
      --# pre N > Types.No_Node;
   is
      The_Element : Types.Node_Id;
   begin
      The_Element := Nodes.A_Element (N);
      Stack.Push (Stack.E_Composite, The_Element);
   end W_Syntactic_Primary;

   -------------------
   -- W_Syntax_Rule --
   -------------------

   --  (* see 4.3 *) syntax rule
   --    = meta identifier, defining symbol,
   --      definitions list, terminator symbol;

   procedure W_Syntax_Rule (N : Types.Node_Id)
      --# global in     Nodes.State;
      --#        in out Stack.State;
      --# pre N > Types.No_Node;
   is
      The_Meta_Identifier  : Types.Node_Id;
      The_Definitions_List : Types.Node_Id;
   begin
      The_Meta_Identifier  := Nodes.Meta_Identifier (N);
      The_Definitions_List := Nodes.Definitions_List (N);

      Stack.Push_Temp (Stack.E_Composite, The_Meta_Identifier);
      Stack.Push_Temp (Stack.E_Defining_Symbol, Types.No_Node);
      Stack.Push_Temp (Stack.E_Composite, The_Definitions_List);
      Stack.Push_Temp (Stack.E_Terminator_Symbol, Types.No_Node);
      Stack.Pump;
   end W_Syntax_Rule;

   --------------
   -- W_Syntax --
   --------------

   --  (* see 4.2 *) syntax
   --    = syntax rule, {syntax rule};

   procedure W_Syntax (N : Types.Node_Id)
      --# global in     Nodes.State;
      --#        in out Stack.State;
      --# pre N > Types.No_Node;
   is
      The_Syntax_Rules : Types.List_Id;
      I                : Types.Node_Id;
   begin
      The_Syntax_Rules := Nodes.Syntax_Rules (N);

      I := Nodes.First_Node (The_Syntax_Rules);
      while I /= Types.No_Node loop
         Stack.Push_Temp (Stack.E_Composite, I);
         I := Nodes.Next_Node (I);
      end loop;
      Stack.Pump;
   end W_Syntax;

   ----------------------
   -- W_Syntactic_Term --
   ----------------------

   --  (* see 4.6 *) syntactic term
   --    = syntactic factor,
   --      [except symbol, syntactic exception];

   procedure W_Syntactic_Term (N : Types.Node_Id)
      --# global in     Nodes.State;
      --#        in out Stack.State;
      --# pre N > Types.No_Node;
   is
      The_Syntactic_Factor : Types.Node_Id;

   --        The_Syntactic_Exception : Types.Node_Id := Types.No_Node;
   begin
      The_Syntactic_Factor := Nodes.Syntactic_Factor (N);
      Stack.Push (Stack.E_Composite, The_Syntactic_Factor);

      --        T := Next_Token;
      --        if T = Lexer.T_Except_Symbol then
      --           B := P_Except_Symbol;
      --           The_Syntactic_Exception := P_Syntactic_Exception;
      --        end if;

   end W_Syntactic_Term;

   -----------
   -- Write --
   -----------

   procedure Write (N : Types.Node_Id)
   --# global in     Lexer.Token_Image;
   --#        in     Namet.State;
   --#        in     Nodes.State;
   --#        in out Output.State;
   --#        in out Stack.State;
   --# pre N > Types.No_Node;
   is
      Node_To_Print     : Types.Node_Id;
      Node_Element_Kind : Stack.Element_Kind_Type;

   begin
      Stack.Push (Stack.E_Composite, N);

      loop
         exit when Stack.Is_Empty;
         Stack.Pop (Node_Element_Kind, Node_To_Print);

         case Node_Element_Kind is
         when Stack.E_Composite =>

            if Node_To_Print > Types.No_Node then
               case Nodes.Kind (Node_To_Print) is
                  when Nodes.K_Syntax =>
                     W_Syntax (Node_To_Print);
                  when Nodes.K_Syntactic_Factor =>
                     W_Syntactic_Factor (Node_To_Print);
                  when Nodes.K_Syntax_Rule =>
                     W_Syntax_Rule (Node_To_Print);
                  when Nodes.K_Definitions_List =>
                     W_Definitions_List (Node_To_Print);
                  when Nodes.K_Meta_Identifier =>
                     W_Meta_Identifier (Node_To_Print);
                  when Nodes.K_Syntactic_Primary =>
                     W_Syntactic_Primary (Node_To_Print);
                  when Nodes.K_Single_Definition =>
                     W_Single_Definition (Node_To_Print);
                  when Nodes.K_Syntactic_Term =>
                     W_Syntactic_Term (Node_To_Print);
                  when Nodes.K_Optional_Sequence =>
                     W_Optional_Sequence (Node_To_Print);
                  when Nodes.K_Repeated_Sequence =>
                     W_Repeated_Sequence (Node_To_Print);
                  when Nodes.K_Grouped_Sequence =>
                     W_Grouped_Sequence (Node_To_Print);
                  when Nodes.K_Terminal_String =>
                     W_Terminal_String (Node_To_Print);
                  when Nodes.K_Empty_Sequence =>
                     W_Empty_Sequence;
                  when others =>
                     null;
               end case;
            else
               null; -- XXX
            end if;

         when Stack.E_Start_Group_Symbol =>
            W_Start_Group_Symbol;
         when Stack.E_End_Group_Symbol =>
            W_End_Group_Symbol;
         when Stack.E_Start_Option_Symbol =>
            W_Start_Option_Symbol;
         when Stack.E_End_Option_Symbol =>
            W_End_Option_Symbol;
         when Stack.E_Start_Repeat_Symbol =>
            W_Start_Repeat_Symbol;
         when Stack.E_End_Repeat_Symbol =>
            W_End_Repeat_Symbol;
         when Stack.E_Concatenate_Symbol =>
            W_Concatenate_Symbol;
         when Stack.E_Defining_Symbol =>
            W_Defining_Symbol;
         when Stack.E_Terminator_Symbol =>
            W_Terminator_Symbol;
         when Stack.E_Definition_Separator_Symbol =>
            W_Definition_Separator_Symbol;
         when others =>
            null;
         end case;
      end loop;
   end Write;

end EBNF.Printer;
