------------------------------------------------------------------------------
--                                                                          --
--                         BlackShield COMPONENTS                           --
--                                                                          --
--                               E R R O R S                                --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                     Copyright (C) 2010, J. Hugues.                       --
--                                                                          --
-- BlackShield is free software;  you  can  redistribute  it and/or  modify --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. BlackShield is distributed  in  the  hope  that it will be--
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with BlackShield; see file COPYING--
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--   BlackShield is maintained by Jerome Hugues <hugues.jerome@gmail.com>   --
--                                                                          --
------------------------------------------------------------------------------

with Output;
with Namet;
with Spark.Command_Line;
with Spark.OS_Lib;
with Spark.Strings;

with Types;              use type Types.Int, Types.Name_Id;
with Locations;          use type Locations.Location;

package body Errors is

   -----------------
   -- Check_Space --
   -----------------

   procedure Check_Space is
   begin
      if Namet.Name_Len > 0
        and then Namet.Name_Buffer (Namet.Name_Len) /= ' '
      then
         Namet.Add_Char_To_Name_Buffer (' ');
      end if;
   end Check_Space;

   ----------------
   -- Initialize --
   ----------------

   procedure Initialize is
   begin
      Error_Loc  := Error_Locs'(Error_Range => Locations.No_Location);
      Error_Name := Error_Names'(Error_Range => Types.No_Name);
      Error_Int  := Error_Ints'(Error_Range => 0);
   end Initialize;

   ------------------------------
   -- Internal_Display_Message --
   ------------------------------

   procedure Internal_Display_Message (Msg_Type : Message_Type; S : String) is
      I       : Error_Range := 1;
      N       : Error_Range := 1;
      M       : Boolean := False; --  Meta-character
      J       : Integer;
      Loc_Img : Spark.Strings.SparkString;
   begin
      if Error_Loc (1) /= Locations.No_Location then
         --  Note: we expect the assignment to Namet.Name_Len to be
         --  ineffective: we are trashing the value of Namet.Name_Len
         --  after the call to Location_Image.

         --# accept F, 10, Namet.Name_Len,
         --#  "Expected ineffective assignment to Namet.Name_Len";
         Locations.Location_Image (Error_Loc (1), Loc_Img);
         Namet.Set_SparkStr_To_Name_Buffer (Loc_Img);
      end if;

      Namet.Add_Str_To_Name_Buffer (": ");

      if S'Last >= S'First then
         J := S'First;
         loop
            --# assert J >= S'First
            --# and J <= S'Last and I >= 1 and I <= 2;

            if S (J) = '$' then
               Namet.Add_Nat_To_Name_Buffer (Error_Int (I));
               if I = 1 then
                  I := 2;
               else
                  --   XXX
                  null;
               end if;
               M := False;

            elsif S (J) = '#' then
               Check_Space;
               if Error_Name (N) > Types.No_Name then
                  Namet.Add_Char_To_Name_Buffer ('"'); -- "
                  Namet.Get_Name_String_And_Append (Error_Name (N));
                  Namet.Add_Char_To_Name_Buffer ('"'); -- "
                  if N = 1 then
                     N := 2;
                  else
                     --   XXX
                     null;
                  end if;
               else
                  --   XXX
                  null;
               end if;
               M := True;

            else
               if M then
                  Namet.Add_Char_To_Name_Buffer (' ');
                  M := False;
               end if;
               Namet.Add_Char_To_Name_Buffer (S (J));
            end if;

            exit when J = S'Last;
            J := J + 1;
         end loop;
      else
         Namet.Set_Str_To_Name_Buffer ("ICE");
      end if;

      Output.Set_Standard_Error;
      case Msg_Type is
         when Notice =>
            null;
         when Warning =>
            Output.Write_Str ("Warning: ");
         when Error =>
            Output.Write_Str ("Error: ");
      end case;

      Output.Write_SparkStr (Namet.Name_Buffer_Content);
      Output.Write_Eol;
      Output.Set_Standard_Output;

      Initialize;
   end Internal_Display_Message;

   -------------------
   -- Display_Error --
   -------------------

   procedure Display_Error (S : String) is
   begin
      if N_Errors < 10 then
         N_Errors := N_Errors + 1;
      end if;

      Internal_Display_Message (Error, S);

      if N_Errors = 10 then
         Spark.OS_Lib.OS_Exit (1);
      end if;
   end Display_Error;

   ---------------------
   -- Display_Warning --
   ---------------------

   procedure Display_Warning (S : String) is
   begin
      if N_Warnings < 10 then
         N_Warnings := N_Warnings + 1;
      end if;

      Internal_Display_Message (Warning, S);

      if N_Warnings = 10 then
         Spark.OS_Lib.OS_Exit (1);
      end if;
   end Display_Warning;

   --------------------
   -- Display_Status --
   --------------------

   procedure Display_Status is
   begin
      if N_Errors > 0 then
         Error_Int (1) := N_Errors;
         Error_Int (2) := N_Warnings;

         if N_Warnings > 0 then
            Display_Error ("$ error(s) and $ warning(s)");
         else
            Display_Error ("$ error(s)");
         end if;
         Spark.Command_Line.Set_Exit_Status (1);

      elsif N_Warnings > 0 then
         Error_Int (1) := N_Warnings;
         Display_Error ("$ warning(s)");
         Spark.Command_Line.Set_Exit_Status (0);

      else
         --  Nothing to report
         Spark.Command_Line.Set_Exit_Status (0);
      end if;
   end Display_Status;

   -------------------
   -- Set_Error_Loc --
   -------------------

   procedure Set_Error_Loc
     (Position      : Error_Range;
      The_Error_Loc : Locations.Location)
   is
   begin
      Error_Loc (Position) := The_Error_Loc;
   end Set_Error_Loc;

   --------------------
   -- Set_Error_Name --
   --------------------

   procedure Set_Error_Name
     (Position       : Error_Range;
      The_Error_Name : Types.Name_Id)
   is
   begin
      Error_Name (Position) := The_Error_Name;
   end Set_Error_Name;

end Errors;
