------------------------------------------------------------------------------
--                                                                          --
--                         BlackShield COMPONENTS                           --
--                                                                          --
--                                N A M E T                                 --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                     Copyright (C) 2010, J. Hugues.                       --
--                                                                          --
-- BlackShield is free software;  you  can  redistribute  it and/or  modify --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. BlackShield is distributed  in  the  hope  that it will be--
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with BlackShield; see file COPYING--
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--   BlackShield is maintained by Jerome Hugues <hugues.jerome@gmail.com>   --
--                                                                          --
------------------------------------------------------------------------------

with Output;
with Ada.Characters.Latin_1;
with Spark.Type_Image;

package body Namet
   --# own State is Name_Chars_Table, Name_Chars_Last, Name_Entries_Table,
   --# Name_Entries_Last, Hash_Table;
is

   --  This table stores the actual string names. Although logically there
   --  is no need for a terminating character (since the length is stored
   --  in the name entry table), we still store a NUL character at the end
   --  of every name (for convenience in interfacing to the C world).

   Max_Name_Chars_Id : constant := 50_000;
   subtype Name_Chars_Table_Range is Types.Nat range 0 .. Max_Name_Chars_Id;
   type Name_Chars_Table_Type is array (Name_Chars_Table_Range) of Character;
   Name_Chars_Table : Name_Chars_Table_Type  :=
     Name_Chars_Table_Type'(Name_Chars_Table_Range => ' ');
   Name_Chars_Last  : Name_Chars_Table_Range := Name_Chars_Table'First;

   type Name_Entry is record
      Name_Chars_Index : Name_Chars_Table_Range;
      --  Starting location of characters in the Name_Chars table minus
      --  one (i.e. pointer to character just before first character). The
      --  reason for the bias of one is that indexes in Name_Buffer are
      --  one's origin, so this avoids unnecessary adds and subtracts of 1.

      Name_Len : Name_Len_Type;
      --  Length of this name in characters

      Byte_Info : Types.Byte;
      --  Byte value associated with this name

      Hash_Link : Types.Name_Id;
      --  Link to next entry in names table for same hash code

      Int_Info : Types.Int;
      --  Int Value associated with this name
   end record;

   Default_Name_Entry : constant Name_Entry :=
      Name_Entry'(0, 0, 0, Types.No_Name, 0);

   --  This is the table that is referenced by Types.Name_Id entries.
   --  It contains one entry for each unique name in the table.

   subtype Name_Entries_Table_Range is Types.Name_Id range
      Types.First_Name_Id .. Types.Name_Id'Last;
   type Name_Entries_Table_Type is
     array (Name_Entries_Table_Range) of Name_Entry;
   Name_Entries_Table : Name_Entries_Table_Type :=
      Name_Entries_Table_Type'
     (Name_Entries_Table_Range => Default_Name_Entry);
   Name_Entries_Last  : Types.Name_Id           := Types.First_Name_Id - 1;

   type Hash_Table_Type is array (Hash_Index_Type) of Types.Name_Id;
   Hash_Table : Hash_Table_Type :=
      Hash_Table_Type'(Hash_Index_Type => Types.No_Name);

   --  The hash table is used to locate existing entries in the names table.
   --  The entries point to the first names table entry whose hash value
   --  matches the hash code. Then subsequent names table entries with the
   --  same hash code value are linked through the Hash_Link fields.

   -------------------------------
   -- Name_Chars_Increment_Last --
   -------------------------------

   procedure Name_Chars_Increment_Last
   --# global in out Name_Chars_Last;
   --# post Name_Chars_Last = Name_Chars_Last~ + 1
   --#   or Name_Chars_Last = Max_Name_Chars_Id;
   --   XXX does not accept Name_Chars_Table'Last
   is
   begin
      if Name_Chars_Last < Max_Name_Chars_Id then
         Name_Chars_Last := Name_Chars_Last + 1;
      else
         --  XXX should do something
         null;
      end if;
   end Name_Chars_Increment_Last;

   ---------------------------------
   -- Name_Entries_Increment_Last --
   ---------------------------------

   procedure Name_Entries_Increment_Last
   --# global in out Name_Entries_Last;
   --# post (Name_Entries_Last = Types.Name_Id'Last
   --#       or Name_Entries_Last = Name_Entries_Last~ + 1)
   --#  and Name_Entries_Last > Types.No_Name;
   is
   begin
      if Name_Entries_Last < Types.Name_Id'Last then
         Name_Entries_Last := Name_Entries_Last + 1;
      else
         --  XXX should do something
         null;
      end if;
   end Name_Entries_Increment_Last;

   -------------------------
   -- Name_Buffer_Content --
   -------------------------

   function Name_Buffer_Content return  Spark.Strings.SparkString is
      Result  : Spark.Strings.SparkString := Spark.Strings.EmptyString;
      Success : Boolean;
   begin
      for J in Positive range 1 .. Name_Len loop
         Spark.Strings.AppendChar (Result, Name_Buffer (J), Success);
         exit when not Success;
      end loop;

      return Result;
   end Name_Buffer_Content;

   -----------------------------
   -- Add_Char_To_Name_Buffer --
   -----------------------------

   procedure Add_Char_To_Name_Buffer (C : Character) is
   begin
      if Name_Len < Name_Buffer'Last then
         Name_Len               := Name_Len + 1;
         Name_Buffer (Name_Len) := C;
      end if;
   end Add_Char_To_Name_Buffer;

   ----------------------------
   -- Add_Nat_To_Name_Buffer --
   ----------------------------

   procedure Add_Nat_To_Name_Buffer (V : Types.Nat) is
      V_Image : Spark.Strings.SparkString;
   begin
      V_Image := Spark.Type_Image.Integer_Image (Integer (V));
      --# assert V_Image.Length <= 255;

      for J in Positive range 1 .. V_Image.Length loop
         --# assert J in 1 .. V_Image.Length
         --#    and V_Image = V_Image%;

         Add_Char_To_Name_Buffer (V_Image.Text (J));
      end loop;
   end Add_Nat_To_Name_Buffer;

   ----------------------------
   -- Add_Str_To_Name_Buffer --
   ----------------------------

   procedure Add_Str_To_Name_Buffer (S : String) is
   begin
      for J in Integer range S'Range loop
         Add_Char_To_Name_Buffer (S (J));
      end loop;
   end Add_Str_To_Name_Buffer;

   ---------------------------------
   -- Add_SparkStr_To_Name_Buffer --
   ---------------------------------

   procedure Add_SparkStr_To_Name_Buffer (S : Spark.Strings.SparkString) is
   begin
      for J in Positive range 1 .. S.Length loop
         Add_Char_To_Name_Buffer (S.Text (J));
      end loop;
   end Add_SparkStr_To_Name_Buffer;

   ---------------------
   -- Get_Name_String --
   ---------------------

   procedure Get_Name_String (Id : Types.Name_Id)
   --# global in     Name_Chars_Table;
   --#        in     Name_Entries_Table;
   --#        in out Name_Buffer;
   --#           out Name_Len;
   --# pre Id > Types.No_Name;
   is
      S : Types.Int;
      J : Types.Int;

   begin
      S := Name_Entries_Table (Id).Name_Chars_Index;
      --# assert S >= 0
      --#    and S <= Name_Chars_Table'Last;

      Name_Len := Name_Entries_Table (Id).Name_Len;

      if S + Types.Int (Name_Len) > Name_Chars_Table'Last then
         --  XXX
         null;
      else
         --# assert S >= 0
         --#    and S <= Name_Chars_Table'Last
         --#    and S + Types.Int (Name_Len) <= Name_Chars_Table'Last
         --#    and Name_Len >= 0 and Name_Len <= Name_Buffer_Max;

         J := 1;
         loop
         --# assert J >= 1
         --#    and S >= 0
         --#    and S <= Name_Chars_Table'Last
         --#    and S + Types.Int (Name_Len) <= Name_Chars_Table'Last
         --#    and Name_Len >= 0
         --#    and Name_Len <= Name_Buffer_Max;

            exit when J > Types.Int (Name_Len);
            --# assert J >= 1
            --#    and S >= 0
            --#    and S <= Name_Chars_Table'Last
            --#    and J <= Types.Int (Name_Len)
            --#    and S + Types.Int (Name_Len) <= Name_Chars_Table'Last
            --#    and Name_Len >= 0
            --#    and Name_Len <= Name_Buffer_Max;

            Name_Buffer (Integer (J))  := Name_Chars_Table (S + J);
            J                          := J + 1;
         end loop;
      end if;
   end Get_Name_String;

   -----------------
   -- Name_String --
   -----------------

   function Name_String
     (Id   : Types.Name_Id)
      return Spark.Strings.SparkString
   --# global in Name_Chars_Table;
   --#        in Name_Entries_Table;
   --# pre Id > Types.No_Name;
   is
      First_Index, Last_Index, J : Types.Int;
      Result                     : Spark.Strings.SparkString :=
        Spark.Strings.EmptyString;
      Success                    : Boolean;
   begin
      First_Index := Name_Entries_Table (Id).Name_Chars_Index + 1;
      --# assert First_Index > 0;

      Last_Index := Name_Entries_Table (Id).Name_Chars_Index +
                    Types.Int (Name_Entries_Table (Id).Name_Len);

      if First_Index > Name_Chars_Table'Last
        or else Last_Index > Name_Chars_Table'Last
      then
         Result := Spark.Strings.EmptyString;
      else
         --# assert First_Index > 0
         --#   and First_Index <= Name_Chars_Table'Last
         --#   and Last_Index <= Name_Chars_Table'Last;

         J := First_Index;
         loop
         --# assert J >= First_Index
         --#    and First_Index > 0
         --#    and First_Index <= Name_Chars_Table'Last
         --#    and Last_Index <= Name_Chars_Table'Last;

            exit when J > Last_Index;
            --# assert J <= Last_Index
            --#    and J >= First_Index
            --#    and First_Index > 0
            --#    and First_Index <= Name_Chars_Table'Last
            --#    and Last_Index <= Name_Chars_Table'Last;

            Spark.Strings.AppendChar (Result, Name_Chars_Table (J), Success);

            if not Success then
               Result := Spark.Strings.EmptyString;
               exit;
            end if;
            J := J + 1;
         end loop;
      end if;

      return Result;
   end Name_String;

   --------------------------------
   -- Get_Name_String_And_Append --
   --------------------------------

   procedure Get_Name_String_And_Append (Id : Types.Name_Id)
   --# global in     Name_Chars_Table;
   --#        in     Name_Entries_Table;
   --#        in out Name_Buffer;
   --#        in out Name_Len;
   --# pre Id > Types.No_Name;
   is
      S : Types.Int;
      J : Types.Int;

   begin
      S := Name_Entries_Table (Id).Name_Chars_Index;
      --# assert S >= 0
      --#    and S <= Name_Chars_Table'Last;

      if S + Types.Int (Name_Entries_Table (Id).Name_Len) >
         Name_Chars_Table'Last
        or else S + Types.Int (Name_Len) >= Types.Int (Name_Buffer'Last)
      then
         null;
      else
         --# assert S + Types.Int (Name_Entries_Table (Id).Name_Len)
         --#        <= Name_Chars_Table'Last
         --#    and S <= Name_Chars_Table'Last
         --#    and S >= 0;

         J := 1;
         loop
         --# assert J >= 1
         --#    and S + Types.Int (Name_Entries_Table (Id).Name_Len)
         --#        <= Name_Chars_Table'Last
         --#    and S <= Name_Chars_Table'Last
         --#    and S >= 0;

            exit when J > Types.Int (Name_Entries_Table (Id).Name_Len)
                     or else Name_Len = Name_Len_Type'Last;

            --# assert Name_Len < Name_Len_Type'Last
            --#    and J <= Types.Int (Name_Entries_Table (Id).Name_Len)
            --#    and S + Types.Int (Name_Entries_Table (Id).Name_Len)
            --#        <= Name_Chars_Table'Last
            --#    and S <= Name_Chars_Table'Last
            --#    and S >= 0 and J >= 1;

            Name_Len               := Name_Len + 1;
            Name_Buffer (Name_Len) := Name_Chars_Table (S + J);

            J := J + 1;
         end loop;
      end if;
   end Get_Name_String_And_Append;

   -------------------------
   -- Get_Name_Table_Byte --
   -------------------------

   function Get_Name_Table_Byte (Id : Types.Name_Id) return Types.Byte
   --# global in Name_Entries_Table;
   --# pre Id > Types.No_Name;
   is
   begin
      return Name_Entries_Table (Id).Byte_Info;
   end Get_Name_Table_Byte;

   -------------------------
   -- Get_Name_Table_Info --
   -------------------------

   function Get_Name_Table_Info (Id : Types.Name_Id) return Types.Int
   --# global in Name_Entries_Table;
   --# pre Id > Types.No_Name;
   is
   begin
      return Name_Entries_Table (Id).Int_Info;
   end Get_Name_Table_Info;

   ----------
   -- Hash --
   ----------

   function Hash return Hash_Index_Type is
      subtype Int_1_12 is Types.Int range 1 .. 12;
      --  Used to avoid when others on case jump below

      Even_Name_Len : Integer;
      --  Last even numbered position (used for >12 case)
      Result : Hash_Index_Type;
      The_Hash : Integer;

   begin
      --  Special test for 12 (rather than counting on a when others for the
      --  case statement below) avoids some Ada compilers converting the case
      --  statement into successive jumps.

      --  The case of a name longer than 12 characters is handled by taking
      --  the first 6 odd numbered characters and the last 6 even numbered
      --  characters

      if Name_Len > 12 then
         if Name_Len mod 2 = 0 then
            Even_Name_Len := Name_Len;
         else
            Even_Name_Len := Name_Len - 1;
         end if;

         The_Hash :=
           ((((((((((((Character'Pos (Name_Buffer (1))) * 2 +
                      Character'Pos (Name_Buffer (Even_Name_Len - 10))) *
                     2 +
                     Character'Pos (Name_Buffer (3))) *
                    2 +
                    Character'Pos (Name_Buffer (Even_Name_Len - 8))) *
                   2 +
                   Character'Pos (Name_Buffer (5))) *
                  2 +
                  Character'Pos (Name_Buffer (Even_Name_Len - 6))) *
                 2 +
                 Character'Pos (Name_Buffer (7))) *
                2 +
                Character'Pos (Name_Buffer (Even_Name_Len - 4))) *
               2 +
               Character'Pos (Name_Buffer (9))) *
              2 +
              Character'Pos (Name_Buffer (Even_Name_Len - 2))) *
             2 +
             Character'Pos (Name_Buffer (11))) *
            2 +
              Character'Pos (Name_Buffer (Even_Name_Len)));

         Result := Types.Int (The_Hash mod Hash_Num);

      elsif Name_Len = 0 then
         Result := 0;

      else
         --  For the cases of 1-12 characters, all characters
         --  participate in the hash. The positioning is randomized,
         --  with the bias that characters later on participate fully
         --  (i.e. are added towards the right side).

         case Int_1_12 (Name_Len) is
         when 1 =>
            Result := Character'Pos (Name_Buffer (1));

         when 2 =>
            Result :=
              ((Character'Pos (Name_Buffer (1))) * 64 +
               Character'Pos (Name_Buffer (2))) mod
              Hash_Num;

         when 3 =>
            Result :=
              (((Character'Pos (Name_Buffer (1))) * 16 +
                Character'Pos (Name_Buffer (3))) *
               16 +
               Character'Pos (Name_Buffer (2))) mod
              Hash_Num;

         when 4 =>
            Result :=
              ((((Character'Pos (Name_Buffer (1))) * 8 +
                 Character'Pos (Name_Buffer (2))) *
                8 +
                Character'Pos (Name_Buffer (3))) *
               8 +
               Character'Pos (Name_Buffer (4))) mod
              Hash_Num;

         when 5 =>
            Result :=
              (((((Character'Pos (Name_Buffer (4))) * 8 +
                  Character'Pos (Name_Buffer (1))) *
                 4 +
                 Character'Pos (Name_Buffer (3))) *
                4 +
                Character'Pos (Name_Buffer (5))) *
               8 +
               Character'Pos (Name_Buffer (2))) mod
              Hash_Num;

         when 6 =>
            The_Hash :=
              ((((((Character'Pos (Name_Buffer (5))) * 4 +
                     Character'Pos (Name_Buffer (1))) *
                    4 +
                    Character'Pos (Name_Buffer (4))) *
                   4 +
                   Character'Pos (Name_Buffer (2))) *
                  4 +
                  Character'Pos (Name_Buffer (6))) *
                 4 +
                 Character'Pos (Name_Buffer (3)));

            Result := Types.Int (The_Hash mod Hash_Num);

         when 7 =>
            The_Hash :=
              (((((((Character'Pos (Name_Buffer (4))) * 4 +
                    Character'Pos (Name_Buffer (3))) *
                   4 +
                   Character'Pos (Name_Buffer (1))) *
                  4 +
                  Character'Pos (Name_Buffer (2))) *
                 2 +
                 Character'Pos (Name_Buffer (5))) *
                2 +
                Character'Pos (Name_Buffer (7))) *
               2 +
               Character'Pos (Name_Buffer (6)));

            Result := Types.Int (The_Hash mod Hash_Num);

         when 8 =>
            The_Hash :=
              ((((((((Character'Pos (Name_Buffer (2))) * 4 +
                     Character'Pos (Name_Buffer (1))) *
                    4 +
                    Character'Pos (Name_Buffer (3))) *
                   2 +
                   Character'Pos (Name_Buffer (5))) *
                  2 +
                  Character'Pos (Name_Buffer (7))) *
                 2 +
                 Character'Pos (Name_Buffer (6))) *
                2 +
                Character'Pos (Name_Buffer (4))) *
               2 +
                 Character'Pos (Name_Buffer (8)));

            Result := Types.Int (The_Hash mod Hash_Num);

         when 9 =>
            The_Hash :=
              (((((((((Character'Pos (Name_Buffer (2))) * 4 +
                      Character'Pos (Name_Buffer (1))) *
                     4 +
                     Character'Pos (Name_Buffer (3))) *
                    4 +
                    Character'Pos (Name_Buffer (4))) *
                   2 +
                   Character'Pos (Name_Buffer (8))) *
                  2 +
                  Character'Pos (Name_Buffer (7))) *
                 2 +
                 Character'Pos (Name_Buffer (5))) *
                2 +
                Character'Pos (Name_Buffer (6))) *
               2 +
               Character'Pos (Name_Buffer (9)));

            Result := Types.Int (The_Hash mod Hash_Num);

         when 10 =>
            The_Hash :=
              ((((((((((Character'Pos (Name_Buffer (01))) * 2 +
                       Character'Pos (Name_Buffer (02))) *
                      2 +
                      Character'Pos (Name_Buffer (08))) *
                     2 +
                     Character'Pos (Name_Buffer (03))) *
                    2 +
                    Character'Pos (Name_Buffer (04))) *
                   2 +
                   Character'Pos (Name_Buffer (09))) *
                  2 +
                  Character'Pos (Name_Buffer (06))) *
                 2 +
                 Character'Pos (Name_Buffer (05))) *
                2 +
                Character'Pos (Name_Buffer (07))) *
               2 +
                 Character'Pos (Name_Buffer (10)));

            Result := Types.Int (The_Hash mod Hash_Num);

         when 11 =>
            The_Hash :=
              (((((((((((Character'Pos (Name_Buffer (05))) * 2 +
                        Character'Pos (Name_Buffer (01))) *
                       2 +
                       Character'Pos (Name_Buffer (06))) *
                      2 +
                      Character'Pos (Name_Buffer (09))) *
                     2 +
                     Character'Pos (Name_Buffer (07))) *
                    2 +
                    Character'Pos (Name_Buffer (03))) *
                   2 +
                   Character'Pos (Name_Buffer (08))) *
                  2 +
                  Character'Pos (Name_Buffer (02))) *
                 2 +
                 Character'Pos (Name_Buffer (10))) *
                2 +
                Character'Pos (Name_Buffer (04))) *
               2 +
               Character'Pos (Name_Buffer (11)));

            Result := Types.Int (The_Hash mod Hash_Num);

            when 12 =>
               The_Hash :=
              ((((((((((((Character'Pos (Name_Buffer (03))) * 2 +
                         Character'Pos (Name_Buffer (02))) *
                        2 +
                        Character'Pos (Name_Buffer (05))) *
                       2 +
                       Character'Pos (Name_Buffer (01))) *
                      2 +
                      Character'Pos (Name_Buffer (06))) *
                     2 +
                     Character'Pos (Name_Buffer (04))) *
                    2 +
                    Character'Pos (Name_Buffer (08))) *
                   2 +
                   Character'Pos (Name_Buffer (11))) *
                  2 +
                  Character'Pos (Name_Buffer (07))) *
                 2 +
                 Character'Pos (Name_Buffer (09))) *
                2 +
                Character'Pos (Name_Buffer (10))) *
               2 +
                 Character'Pos (Name_Buffer (12)));

            Result := Types.Int (The_Hash mod Hash_Num);
         end case;
      end if;

      return Result;
   end Hash;

   ----------------
   -- Initialize --
   ----------------

   procedure Initialize
   --# global in out Hash_Table;
   --#        in out Name_Chars_Table;
   --#        in out Name_Entries_Table;
   --#           out Name_Chars_Last;
   --#           out Name_Entries_Last;
   is
   begin
      --  Initialize entries for one character names

      Name_Entries_Last := Types.No_Name;
      Name_Chars_Last   := Name_Chars_Table'First;

      for J in Natural range 0 .. 255 loop
         --# assert J >= 0
         --#    and J <= 255
         --#    and Name_Chars_Last in Name_Chars_Table_Range;

         --  XXX investigate why SPARK cannot prove when using Character,
         --  issue with Succ attribute

         Name_Entries_Increment_Last;
         Name_Entries_Table (Name_Entries_Last).Name_Chars_Index :=
           Name_Chars_Last;
         Name_Entries_Table (Name_Entries_Last).Name_Len         := 1;
         Name_Entries_Table (Name_Entries_Last).Hash_Link        :=
           Types.No_Name;
         Name_Entries_Table (Name_Entries_Last).Int_Info         := 0;
         Name_Entries_Table (Name_Entries_Last).Byte_Info        := 0;
         Name_Chars_Increment_Last;
         Name_Chars_Table (Name_Chars_Last) := Character'Val (J);
         Name_Chars_Increment_Last;
         Name_Chars_Table (Name_Chars_Last) := Ada.Characters.Latin_1.NUL;
      end loop;

      --  Clear hash table

      for J in Hash_Index_Type loop
         Hash_Table (J) := Types.No_Name;
      end loop;
   end Initialize;

   ---------------
   -- Name_Find --
   ---------------

   procedure Name_Find (Result : out Types.Name_Id)
   --# global in     Name_Buffer;
   --#        in     Name_Len;
   --#        in out Hash_Table;
   --#        in out Name_Chars_Last;
   --#        in out Name_Chars_Table;
   --#        in out Name_Entries_Last;
   --#        in out Name_Entries_Table;
   --# post Result > Types.No_Name;
   is
      New_Id : Types.Name_Id;
      --  Id of entry in hash search, and value to be returned

      S : Name_Chars_Table_Range;
      --  Pointer into string table

      Hash_Index : Hash_Index_Type;
      --  Computed hash index

      Matched : Boolean := False;
      Exit_Search : Boolean := False;
      J : Positive;

   begin
      Result := Types.No_Name;

      --  Quick handling for one character names

      if Name_Len = 0 then
         --  XXX
         Result := Types.Name_Id'Last;

      elsif Name_Len = 1 then
         Result :=
           Types.Name_Id'(Types.First_Name_Id +
                          Character'Pos (Name_Buffer (1)));

      --  Otherwise search hash table for existing matching entry

      elsif Name_Len > 1 then
         --# assert Name_Len > 1 and not Matched;
         Hash_Index := Hash;
         New_Id     := Hash_Table (Hash_Index);

         if New_Id = Types.No_Name then
            if Name_Entries_Last < Types.Name_Id'Last then
               Hash_Table (Hash_Index) := Name_Entries_Last + 1;
            else
               --  XXX
               null;
            end if;
         else
            --# assert New_Id > Types.No_Name
            --#    and Name_Len > 1 and not Matched;
            Search :
            loop
               --# assert New_Id > Types.No_Name
               --#    and Name_Len > 1 and not Matched;

               if Name_Len = Name_Entries_Table (New_Id).Name_Len then
                  S       := Name_Entries_Table (New_Id).Name_Chars_Index;
                  Matched := True;

                  if S + Types.Int (Name_Len) > Name_Chars_Table'Last then
                     Matched := False;

                  else
                     --# assert New_Id > Types.No_Name
                     --#    and Name_Len > 1
                     --#    and S + Types.Int (Name_Len)
                     --#        <= Name_Chars_Table_Range'Last;

                     J := 1;
                     loop
                        --# assert New_Id > Types.No_Name
                        --#    and Name_Len > 1
                        --#    and J >= 1
                        --#    and J <= Name_Len + 1
                        --#    and S + Types.Int (Name_Len)
                        --#        <= Name_Chars_Table_Range'Last;

                        exit when not Matched
                          or else J > Name_Len
                          or else S + Types.Int (J)
                          > Name_Chars_Table_Range'Last;

                        --# assert New_Id > Types.No_Name
                        --#    and Name_Len > 1
                        --#    and J >= 1
                        --#    and J <= Name_Len
                        --#    and S + Types.Int (Name_Len)
                        --#        <= Name_Chars_Table_Range'Last;

                        if Name_Chars_Table (S + Types.Int (J)) /=
                          Name_Buffer (J)
                        then
                           Matched := False;
                        end if;

                        J := J + 1;
                     end loop;
                  end if;

                  if Matched then
                     Result := New_Id;
                  end if;
               end if;

               --# assert New_Id > Types.No_Name
               --#    and Name_Len > 1
               --#    and ((Matched and Result > Types.No_Name)
               --#        or not Matched);
               --  Current entry in hash chain does not match

               if Name_Entries_Table (New_Id).Hash_Link /=
                 Types.No_Name
               then
                  New_Id := Name_Entries_Table (New_Id).Hash_Link;
               else
                  --# accept F, 41, "Stable expression here expected and OK";
                  if Name_Entries_Last < Types.Name_Id'Last then
                     Name_Entries_Table (New_Id).Hash_Link
                       := Name_Entries_Last + 1;
                  else
                     --  XXX
                     null;
                  end if;
                  Exit_Search := True;

               end if;
               exit Search when Matched
                               or else Exit_Search;
            end loop Search;
         end if;

         --  We fall through here only if a matching entry was not
         --  found in the hash table. We now create a new entry in the
         --  names table. The hash link pointing to the new entry
         --  (Name_Entries_Last+1) has been set.

         --# assert (Matched and Result > Types.No_Name)
         --#    or not Matched;

         if not Matched then
            Name_Entries_Increment_Last;
            --# assert Name_Entries_Last > Types.No_Name;
            Name_Entries_Table (Name_Entries_Last).Name_Chars_Index :=
              Name_Chars_Last;
            Name_Entries_Table (Name_Entries_Last).Name_Len         :=
              Name_Len;
            Name_Entries_Table (Name_Entries_Last).Hash_Link        :=
              Types.No_Name;
            Name_Entries_Table (Name_Entries_Last).Int_Info         := 0;
            Name_Entries_Table (Name_Entries_Last).Byte_Info        := 0;

            --  Set corresponding string entry in the Name_Chars table

            for K in Positive range 1 .. Name_Len loop
               --# assert K in 1 .. Name_Len
               --# and Name_Entries_Last > Types.No_Name;
               Name_Chars_Increment_Last;
               Name_Chars_Table (Name_Chars_Last) := Name_Buffer (K);
            end loop;

            Name_Chars_Increment_Last;
            Name_Chars_Table (Name_Chars_Last) := Ada.Characters.Latin_1.NUL;

            Result := Name_Entries_Last;
            --# assert Result > Types.No_Name;
         end if;
      end if;
   end Name_Find;

   -------------------------
   -- Set_Name_Table_Byte --
   -------------------------

   procedure Set_Name_Table_Byte (Id : Types.Name_Id; Val : Types.Byte)
   --# global in out Name_Entries_Table;
   --# pre Id > Types.No_Name;
   is
   begin
      Name_Entries_Table (Id).Byte_Info := Val;
   end Set_Name_Table_Byte;

   -------------------------
   -- Set_Name_Table_Info --
   -------------------------

   procedure Set_Name_Table_Info (Id : Types.Name_Id; Val : Types.Int)
   --# global in out Name_Entries_Table;
   --# pre Id > Types.No_Name;
   is
   begin
      Name_Entries_Table (Id).Int_Info := Val;
   end Set_Name_Table_Info;

   ----------------------------
   -- Set_Str_To_Name_Buffer --
   ----------------------------

   procedure Set_Str_To_Name_Buffer (S : String) is
   begin
      Name_Len := 0;
      Add_Str_To_Name_Buffer (S);
   end Set_Str_To_Name_Buffer;

   ---------------------------------
   -- Set_SparkStr_To_Name_Buffer --
   ---------------------------------

   procedure Set_SparkStr_To_Name_Buffer (S : Spark.Strings.SparkString) is
   begin
      Name_Len := 0;
      for J in Positive range 1 .. S.Length loop
         Add_Char_To_Name_Buffer (S.Text (J));
      end loop;
   end Set_SparkStr_To_Name_Buffer;

   ----------------
   -- Write_Name --
   ----------------

   procedure Write_Name (Id : Types.Name_Id)
   --# global in     Name_Chars_Table;
   --#        in     Name_Entries_Table;
   --#        in out Name_Buffer;
   --#        in out Output.State;
   --#           out Name_Len;
   --# pre Id > Types.No_Name;
   is
   begin
      Get_Name_String (Id);
      for J in Positive range 1 .. Name_Len loop
         Output.Write_Char (Name_Buffer (J));
      end loop;
   end Write_Name;

end Namet;
