------------------------------------------------------------------------------
--                                                                          --
--                         BlackShield COMPONENTS                           --
--                                                                          --
--                                N A M E T                                 --
--                                                                          --
--                                 S p e c                                  --
--                                                                          --
--                     Copyright (C) 2010, J. Hugues.                       --
--                                                                          --
-- BlackShield is free software;  you  can  redistribute  it and/or  modify --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. BlackShield is distributed  in  the  hope  that it will be--
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with BlackShield; see file COPYING--
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--   BlackShield is maintained by Jerome Hugues <hugues.jerome@gmail.com>   --
--                                                                          --
------------------------------------------------------------------------------

with Types; use type Types.Int, Types.Name_Id;
with Spark.Strings;
--# inherit Ada.Characters.Latin_1,
--#         Output,
--#         Spark.Strings,
--#         Spark.Type_Image,
--#         Types;

package Namet
   --# own Name_Buffer,
   --#     Name_Len,
   --#     State;
   --# initializes Name_Buffer,
   --#             Name_Len,
   --#             State;
is

   --  This package contains routines for handling the names table. The table
   --  is used to store character strings for identifiers and operator symbols,
   --  as well as other string values such as unit names and file names.

   --  The names are hashed so that a given name appears only once in
   --  the table, except that names entered with Name_Enter as opposed
   --  to Name_Find are omitted from the hash table.

   --  The first 26 entries in the names table (with Name_Id values in
   --  the range First_Name_Id .. First_Name_Id + 25) represent names
   --  which are the one character lower case letters in the range
   --  a-z, and these names are created and initialized by the
   --  Initialize procedure.

   --  Two values, one of type Int and one of type Byte, are stored
   --  with each names table entry and subprograms are provided for
   --  setting and retrieving these associated values. The usage of
   --  these values is up to the client.

   Name_Buffer_Max : constant := 16 * 1024;
   subtype Name_Buffer_Range is Natural range 1 .. Name_Buffer_Max;
   subtype Name_Buffer_Str is String (Name_Buffer_Range);

   Name_Buffer : Name_Buffer_Str :=
      Name_Buffer_Str'(Name_Buffer_Range => ' ');
   --  This buffer is used to set the name to be stored in the table for the
   --  Name_Find call, and to retrieve the name for the Get_Name_String call.
   --  The plus 1 in the length allows for cases of adding ASCII.NUL. The
   --  16K here is intended to be an infinite value that ensures that we
   --  never overflow the buffer (names this long are too absurd to worry!)

   subtype Name_Len_Type is Natural range 0 .. Name_Buffer_Max;
   Name_Len : Name_Len_Type := 0;
   --  Length of name stored in Name_Buffer. Used as an input parameter for
   --  Name_Find, and as an output value by Get_Name_String, or Write_Name.

   function Name_Buffer_Content return  Spark.Strings.SparkString;
   --# global in Name_Buffer;
   --#        in Name_Len;

   -----------------
   -- Subprograms --
   -----------------

   procedure Get_Name_String (Id : Types.Name_Id);
   --# global in     State;
   --#        in out Name_Buffer;
   --#           out Name_Len;
   --# pre Id > Types.No_Name;
   --  Get_Name_String is used to retrieve the string associated with an entry
   --  in the names table. The resulting string is stored in Name_Buffer
   --  and Name_Len is set. It is an error to call Get_Name_String with one
   --  of the special name Id values (No_Name or Error_Name).

   function Name_String
     (Id   : Types.Name_Id)
      return Spark.Strings.SparkString;
   --# global in State;
   --# pre Id > Types.No_Name;
   --  This functional form returns the result as a string without affecting
   --  the contents of either Name_Buffer or Name_Len.

   procedure Get_Name_String_And_Append (Id : Types.Name_Id);
   --# global in     State;
   --#        in out Name_Buffer;
   --#        in out Name_Len;
   --# pre Id > Types.No_Name;
   --  Like Get_Name_String but the resulting characters are appended to
   --  the current contents of the entry stored in Name_Buffer, and Name_Len
   --  is incremented to include the added characters.

   function Get_Name_Table_Byte (Id : Types.Name_Id) return Types.Byte;
   --# global in State;
   --# pre Id > Types.No_Name;
   --  Fetches the Byte value associated with the given name

   function Get_Name_Table_Info (Id : Types.Name_Id) return Types.Int;
   --# global in State;
   --# pre Id > Types.No_Name;
   --  Fetches the Int value associated with the given name

   procedure Initialize;
   --# global in out State;
   --  Initializes the names table, including initializing the first 26
   --  entries in the table (for the 1-character lower case names a-z)
   --  Note that Initialize must not be called if Tree_Read is used.

   procedure Name_Find (Result : out Types.Name_Id);
   --# global in     Name_Buffer;
   --#        in     Name_Len;
   --#        in out State;
   --# post Result > Types.No_Name;
   --  Name_Find is called with a string stored in Name_Buffer whose length
   --  is in Name_Len (i.e. the characters of the name are in subscript
   --  positions 1 to Name_Len in Name_Buffer). It searches the names
   --  table to see if the string has already been stored. If so the Id of
   --  the existing entry is returned. Otherwise a new entry is created with
   --  its Name_Table_Info field set to zero. The contents of Name_Buffer
   --  and Name_Len are not modified by this call.

   procedure Add_Char_To_Name_Buffer (C : Character);
   --# global in out Name_Buffer;
   --#        in out Name_Len;
   --  Add given character to the end of the string currently stored in the
   --  Name_Buffer, incrementing Name_Len.

   procedure Add_Nat_To_Name_Buffer (V : Types.Nat);
   --# global in out Name_Buffer;
   --#        in out Name_Len;
   --  Add decimal representation of given value to the end of the string
   --  currently stored in Name_Buffer, incrementing Name_Len as required.

   procedure Add_Str_To_Name_Buffer (S : String);
   --# global in out Name_Buffer;
   --#        in out Name_Len;
   --  Add characters of string S to the end of the string currently stored
   --  in the Name_Buffer, incrementing Name_Len by the length of the string.

   procedure Add_SparkStr_To_Name_Buffer (S : Spark.Strings.SparkString);
   --# global in out Name_Buffer;
   --#        in out Name_Len;
   --  Add characters of string S to the end of the string currently stored
   --  in the Name_Buffer, incrementing Name_Len by the length of the string.

   procedure Set_Str_To_Name_Buffer (S : String);
   --# global in out Name_Buffer;
   --#           out Name_Len;
   --  Equivalent to Name_Len := 0; followed by Add_Str_To_Name_Buffer (S);

   procedure Set_SparkStr_To_Name_Buffer (S : Spark.Strings.SparkString);
   --# global in out Name_Buffer;
   --#           out Name_Len;

   procedure Set_Name_Table_Info (Id : Types.Name_Id; Val : Types.Int);
   --# global in out State;
   --# pre Id > Types.No_Name;
   --  Sets the Int value associated with the given name

   procedure Set_Name_Table_Byte (Id : Types.Name_Id; Val : Types.Byte);
   --# global in out State;
   --# pre Id > Types.No_Name;
   --  Sets the Byte value associated with the given name

   procedure Write_Name (Id : Types.Name_Id);
   --# global in     State;
   --#        in out Name_Buffer;
   --#        in out Output.State;
   --#           out Name_Len;
   --# pre Id > Types.No_Name;
   --  Write_Name writes the characters of the specified name using
   --  the standard output procedures in package Output. No end of
   --  line is written, just the characters of the name. On return
   --  Name_Buffer and Name_Len are set as for a call to
   --  Get_Name_String. The name is written in encoded form.

private

   Hash_Num : constant := 2 ** 12;
   --  Number of headers in the hash table. Current hash algorithm is closely
   --  tailored to this choice, so it can only be changed if a corresponding
   --  change is made to the hash alogorithm.

   Hash_Max : constant Types.Int := Hash_Num - 1;
   --  Indexes in the hash header table run from 0 to Hash_Num - 1

   subtype Hash_Index_Type is Types.Int range 0 .. Hash_Max;
   --  Range of hash index values

   function Hash return Hash_Index_Type;
   --# global in Name_Buffer;
   --#        in Name_Len;
   --  Compute hash code for name stored in Name_Buffer (length in Name_Len)

   procedure Name_Chars_Increment_Last;
   --# global in out State;

   procedure Name_Entries_Increment_Last;
   --# global in out State;

end Namet;
