------------------------------------------------------------------------------
--                                                                          --
--                         BlackShield COMPONENTS                           --
--                                                                          --
--                               O U T P U T                                --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                     Copyright (C) 2010, J. Hugues.                       --
--                                                                          --
-- BlackShield is free software;  you  can  redistribute  it and/or  modify --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. BlackShield is distributed  in  the  hope  that it will be--
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with BlackShield; see file COPYING--
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--   BlackShield is maintained by Jerome Hugues <hugues.jerome@gmail.com>   --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Characters.Latin_1;
with Spark.OS_Lib;
with Spark.Type_Image;

package body Output
   --# own State is Current_FD, Buffer, Next_Column;
is

   Current_FD : System.Address := System.Null_Address;
   --  File descriptor for current output

   -------------------------
   -- Line Buffer Control --
   -------------------------

   --  Note: the following buffer and column position are maintained by
   --  the subprograms defined in this package, and are not normally
   --  directly modified or accessed by a client. However, a client is
   --  permitted to modify these values, using the knowledge that only
   --  Write_Eol actually generates any output.

   Buffer_Max : constant := 8192;
   subtype Buffer_Range is Integer range 1 .. Buffer_Max + 1;
   subtype Buffer_Str is String (Buffer_Range);

   Buffer : Buffer_Str := Buffer_Str'(Buffer_Range => ' ');
   --  Buffer used to build output line. We do line buffering because it
   --  is needed for the support of the debug-generated-code option (-gnatD).
   --  Historically it was first added because on VMS, line buffering is
   --  needed with certain file formats. So in any case line buffering must
   --  be retained for this purpose, even if other reasons disappear. Note
   --  any attempt to write more output to a line than can fit in the buffer
   --  will be silently ignored.

   subtype Column_Range is Natural range 1 .. Buffer_Max + 1;
   Next_Column : Column_Range := 1;
   --  Column about to be written.

   --------------------
   -- Current_Output --
   --------------------

   function Current_Output return System.Address
   --# global in Current_FD;
   is
   begin
      return Current_FD;
   end Current_Output;

   ---------------------------
   -- Decrement_Indentation --
   ---------------------------

   procedure Decrement_Indentation is
   begin
      if N_Space > Space_Increment then
         N_Space := N_Space - Space_Increment;
      else
         N_Space := Natural'First;
      end if;
   end Decrement_Indentation;

   ------------------
   -- Flush_Buffer --
   ------------------

   procedure Flush_Buffer
   --# global in     Buffer;
   --#        in     Current_FD;
   --#        in out Next_Column;
   --# post Next_Column = 1;
   is
      Len : Natural;
   begin
      Len := Natural'(Next_Column) - 1;
      --# assert Len < Column_Range'Last;

      if Len /= 0 then
         for J in Buffer_Range range 1 .. Len loop
            --# assert J >= 1
            --#    and J <= Len
            --#    and J in Buffer_Range
            --#    and Len < Column_Range'Last
            --#    and Len = Len%;
            Spark.OS_Lib.fputc (Character'Pos (Buffer (J)),
                                Current_Output);

         end loop;
      end if;
      Next_Column := 1;
   end Flush_Buffer;

   ---------------------------
   -- Increment_Indentation --
   ---------------------------

   procedure Increment_Indentation is
   begin
      if N_Space < 75 then
         --   XXX use a better value
         N_Space := N_Space + Space_Increment;
      end if;
   end Increment_Indentation;

   ----------------
   -- Set_Output --
   ----------------

   procedure Set_Output (New_Output : System.Address)
   --# global in     Buffer;
   --#        in out Current_FD;
   --#        in out Next_Column;
   is
   begin
      Flush_Buffer;
      Current_FD := New_Output;
   end Set_Output;

   -------------------------
   -- Set_Space_Increment --
   -------------------------

   procedure Set_Space_Increment (Value : Increment_Type) is
   begin
      Space_Increment := Value;
   end Set_Space_Increment;

   ------------------------
   -- Set_Standard_Error --
   ------------------------

   procedure Set_Standard_Error
   --# global in     Buffer;
   --#        in out Current_FD;
   --#        in out Next_Column;
   is
   begin
      Set_Output (Spark.OS_Lib.stderr);
   end Set_Standard_Error;

   -------------------------
   -- Set_Standard_Output --
   -------------------------

   procedure Set_Standard_Output
   --# global in     Buffer;
   --#        in out Current_FD;
   --#        in out Next_Column;
   is
   begin
      Set_Output (Spark.OS_Lib.stdout);
   end Set_Standard_Output;

   ----------------
   -- Write_Char --
   ----------------

   procedure Write_Char (C : Character)
   --# global in out Buffer;
   --#        in out Next_Column;
   is
   begin
      if Next_Column < Buffer'Length then
         Buffer (Next_Column) := C;
         Next_Column          := Next_Column + 1;
      end if;
   end Write_Char;

   ---------------
   -- Write_Eol --
   ---------------

   procedure Write_Eol
   --# global in     Current_FD;
   --#        in out Buffer;
   --#        in out Next_Column;
   is
   begin
      Flush_Buffer;
      Buffer (Next_Column) := Ada.Characters.Latin_1.LF;
      Next_Column          := Next_Column + 1;
      Flush_Buffer;
   end Write_Eol;

   -----------------------
   -- Write_Indentation --
   -----------------------

   procedure Write_Indentation
   --# global in     N_Space;
   --#        in out Buffer;
   --#        in out Next_Column;
   is
   begin
      for J in Natural range 1 .. N_Space loop
         Write_Char (' ');
      end loop;
   end Write_Indentation;

   ---------------
   -- Write_Str --
   ---------------

   procedure Write_Str (S : String)
   --# global in out Buffer;
   --#        in out Next_Column;
   is
   begin
      for J in Positive range S'Range loop
         Write_Char (S (J));
      end loop;
   end Write_Str;

   --------------------
   -- Write_SparkStr --
   --------------------

   procedure Write_SparkStr (S : Spark.Strings.SparkString)
   --# global in out Buffer;
   --#        in out Next_Column;
   is
   begin
      for J in Positive range 1 .. S.Length loop
         Write_Char (S.Text (J));
      end loop;
   end Write_SparkStr;

   ---------------
   -- Write_Int --
   ---------------

   procedure Write_Int (Val : Types.Int)
   --# global in out Buffer;
   --#        in out Next_Column;
   is
   begin
      Write_SparkStr (Spark.Type_Image.Integer_Image (Integer (Val)));
   end Write_Int;

   ----------------
   -- Write_Line --
   ----------------

   procedure Write_Line (S : String)
   --# global in     Current_FD;
   --#        in out Buffer;
   --#        in out Next_Column;
   is
   begin
      Write_Str (S);
      Write_Eol;
   end Write_Line;

end Output;
