------------------------------------------------------------------------------
--                                                                          --
--                         BlackShield COMPONENTS                           --
--                                                                          --
--                               P A R S E R                                --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                     Copyright (C) 2010, J. Hugues.                       --
--                                                                          --
-- BlackShield is free software;  you  can  redistribute  it and/or  modify --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. BlackShield is distributed  in  the  hope  that it will be--
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with BlackShield; see file COPYING--
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--   BlackShield is maintained by Jerome Hugues <hugues.jerome@gmail.com>   --
--                                                                          --
------------------------------------------------------------------------------

with Spark.Command_Line;
with Spark.OS_Lib;
with Spark.Strings;      use type Spark.Strings.SparkString;

with Errors;
with Help;
with EBNF.Lexer;         use type EBNF.Lexer.Token_Type;
with Namet;
with Types;              use type Types.Int, Types.Name_Id;
with EBNF.Parser;
with EBNF.Printer;
with EBNF.Nodes;
with EBNF2IDL;
with Output;
--# inherit EBNF.Lexer,
--#         EBNF.Nodes,
--#         EBNF.Parser,
--#         EBNF.Parser_Stack,
--#         EBNF.Printer,
--#         EBNF2IDL,
--#         Errors,
--#         Help,
--#         Namet,
--#         Output,
--#         Spark.Command_line,
--#         Spark.OS_Lib,
--#         Spark.Strings,
--#         Types;

--# main_program;

procedure Parser
   --# global in     Errors.N_Warnings;
   --#        in out EBNF.Lexer.Buffer;
   --#        in out EBNF.Lexer.Initialized;
   --#        in out EBNF.Lexer.Token;
   --#        in out EBNF.Lexer.Token_Image;
   --#        in out EBNF.Lexer.Token_Location;
   --#        in out EBNF.Lexer.Token_Name;
   --#        in out EBNF.Nodes.N_Indents;
   --#        in out EBNF.Nodes.State;
   --#        in out EBNF.Parser_Stack.State;
   --#        in out EBNF.Printer.Stack_State;
   --#        in out Errors.N_Errors;
   --#        in out Namet.Name_Buffer;
   --#        in out Namet.Name_Len;
   --#        in out Namet.State;
   --#        in out Output.N_Space;
   --#        in out Output.State;
   --#           out Errors.Error_Int;
   --#           out Errors.Error_Loc;
   --#           out Errors.Error_Name;
   --#           out Output.Space_Increment;

is

   Source_Name : Types.Name_Id := Types.No_Name;
   --  File name of the source file

   N : Types.Node_Id;

   Immediate_Exit : Boolean := False;
   Generate_IDL : Boolean := False;
   Dump_Tree : Boolean := False;

   Help_Flag      : Spark.Strings.SparkString;
   Version_Flag   : Spark.Strings.SparkString;
   IDL_Flag       : Spark.Strings.SparkString;
   Dump_Tree_Flag : Spark.Strings.SparkString;

   N_Arguments  : Natural;

begin
   --  Initialization step

   Namet.Initialize;
   Errors.Initialize;
   Output.Set_Space_Increment (2);
   Output.Set_Output (Spark.OS_Lib.stdout);

   Spark.Strings.CopyString (Help_Flag, "--help");
   Spark.Strings.CopyString (Version_Flag, "--version");
   Spark.Strings.CopyString (IDL_Flag, "--idl");
   Spark.Strings.CopyString (Dump_Tree_Flag, "--tree");

   --  Process command line

   N_Arguments := Spark.Command_Line.Argument_Count;

   --# accept F, 22, "Invariant expression OK here";

   if N_Arguments = 0 then
      Help.Usage;
      Immediate_Exit := True;

   else
      --# assert N_Arguments > 0;

      for J in Positive range 1 .. N_Arguments loop
         --  First, handle the --help and the --version flags for
         --  compatibility with the GNU tools.

         if Spark.Command_Line.Argument (J) = Help_Flag then
            Help.Usage;
            Immediate_Exit := True;

         elsif Spark.Command_Line.Argument (J) = Version_Flag then
            Help.Show_Version;
            Immediate_Exit := True;

         elsif Spark.Command_Line.Argument (J) = IDL_Flag then
            Generate_IDL := True;

         elsif Spark.Command_Line.Argument (J) = Dump_Tree_Flag then
            Dump_Tree := True;

         else
            Namet.Set_SparkStr_To_Name_Buffer
              (Spark.Command_Line.Argument (J));
            Namet.Name_Find (Source_Name);
         end if;
      end loop;
   end if;

   if Source_Name = Types.No_Name and then not Immediate_Exit then
      Errors.Display_Error ("no source file provided");

   elsif Errors.N_Errors = 0 and then not Immediate_Exit then
      EBNF.Lexer.Process (Source_Name);
      EBNF.Parser.Parse (N);

      if Types.Present (N)
        and then not Generate_IDL
      then
         EBNF.Printer.Write (N);
      end if;

      if Generate_IDL then
         EBNF2IDL.Generate (N);
      end if;

      if Types.Present (N)
        and then Dump_Tree
      then
         EBNF.Nodes.W_Node (N);
      end if;

   end if;

   --  Output a summary on errors and warnings

   Errors.Display_Status;
end Parser;
