------------------------------------------------------------------------------
--                                                                          --
--                         BlackShield COMPONENTS                           --
--                                                                          --
--                                T Y P E S                                 --
--                                                                          --
--                                 S p e c                                  --
--                                                                          --
--                     Copyright (C) 2010, J. Hugues.                       --
--                                                                          --
-- BlackShield is free software;  you  can  redistribute  it and/or  modify --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. BlackShield is distributed  in  the  hope  that it will be--
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with BlackShield; see file COPYING--
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--   BlackShield is maintained by Jerome Hugues <hugues.jerome@gmail.com>   --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Characters.Latin_1;
with Spark.Strings;
--# inherit Ada.Characters.Latin_1,
--#         Spark.Strings,
--#         Spark.Type_Image;

package Types is

   -------------------------------
   -- General Use Integer Types --
   -------------------------------

   type Short is range -2 ** 15 .. 2 ** 15 - 1;

   type Int is range -2 ** 31 .. +2 ** 31 - 1;
   --  Signed 32-bit integer

   function Int_Image (N : Int) return Spark.Strings.SparkString;

   subtype Nat is Int range 0 .. Int'Last;
   --  Non-negative Int values

   type Byte is mod 2 ** 8;
   --  8-bit unsigned integer

   function Byte_Image (N : Byte) return Spark.Strings.SparkString;

   EOF : constant Character := Ada.Characters.Latin_1.SUB;
   --  The character SUB (16#1A#) is used in DOS and other systems derived
   --  from DOS (OS/2, NT etc) to signal the end of a text file. Internally
   --  all source files are ended by an EOF character, even on Unix systems.
   --  An EOF character acts as the end of file only as the last character
   --  of a source buffer, in any other position, it is treated as a blank
   --  if it appears between tokens, and as an illegal character otherwise.
   --  This makes life easier dealing with files that originated from DOS,
   --  including concatenated files with interspersed EOF characters.

   -----------------------------------------
   -- Types Used for Text Buffer Handling --
   -----------------------------------------

   --  We can't use type String for text buffers, since we must use the
   --  standard 32-bit integer as an index value, since we count on all
   --  index values being the same size.

   Text_Ptr_Max : constant := 100_000;
   subtype Text_Length is Integer range 1 .. Text_Ptr_Max;
   subtype Text_Ptr is Text_Length range 1 .. Text_Ptr_Max;
   --  Type used for subscripts in text buffer

   subtype Text_Buffer is String (Text_Ptr);
   --  Text buffer used to hold source file or library information file

   -----------------------------
   -- Types for Namet Package --
   -----------------------------

   --  Name_Id values are used to identify entries in the names table. Except
   --  for the special values No_Name, and Error_Name, they are subscript
   --  values for the Names table defined in package Namet.

   --  Note that with only a few exceptions, which are clearly documented, the
   --  type Name_Id should be regarded as a private type. In particular it is
   --  never appropriate to perform arithmetic operations using this type.

   Names_Low_Bound : constant := 300_000_000;
   --  Low bound for name Id values

   Names_High_Bound : constant := Names_Low_Bound + (50_000 + 1);
   --  Maximum number of names that can be allocated is 100 million, which is
   --  in practice infinite and there is no need to check the range.

   type Name_Id is range Names_Low_Bound .. Names_High_Bound;
   --  Type used to identify entries in the names table

   No_Str : constant String := "";

   No_Name : constant Name_Id := Names_Low_Bound;
   --  The special Name_Id value No_Name is used in the parser to indicate
   --  a situation where no name is present (e.g. on a loop or block).

   First_Name_Id : constant Name_Id := Names_Low_Bound + 1;
   --  Subscript of first entry in names table

   Max_Node_Id : constant := 10_000;

   -----------
   -- Nodes --
   -----------

   subtype Node_Id is Int range 0 .. Max_Node_Id;
   No_Node : constant Node_Id := 0;

   function Present (E : Node_Id) return Boolean;
   --# return E /= No_Node;
   --  Return true when E is not No_Node

   function No (E : Node_Id) return Boolean;
   --# return E = No_Node;
   --  Return true when E is No_Node

   subtype List_Id is Node_Id;
   No_List : constant List_Id := 0;

end Types;
